
# Remediation Report (copied from cis_reports.py)
cis_report_data = {
    "1.1": {
        "Description": "To apply least-privilege security principle, one can create service-level administrators in corresponding groups and assigning specific users to each service-level administrative group in a tenancy. This limits administrative access in a tenancy.<br><br>It means service-level administrators can only manage resources of a specific service.<br><br>Example policies for global/tenant level service-administrators\n<pre>\nAllow group VolumeAdmins to manage volume-family in tenancy\nAllow group ComputeAdmins to manage instance-family in tenancy\nAllow group NetworkAdmins to manage virtual-network-family in tenancy\n</pre>\nOrganizations have various ways of defining service-administrators. Some may prefer creating service administrators at a tenant level and some per department or per project or even per application environment (dev/test/production etc.). Either approach works so long as the policies are written to limit access given to the service-administrators.<br><br>Example policies for compartment level service-administrators <br><br><pre>Allow group NonProdComputeAdmins to manage instance-family in compartment dev\nAllow group ProdComputeAdmins to manage instance-family in compartment production\nAllow group A-Admins to manage instance-family in compartment Project-A\nAllow group A-Admins to manage volume-family in compartment Project-A\n</pre>",
        "Rationale": "Creating service-level administrators helps in tightly controlling access to Oracle Cloud Infrastructure (OCI) services to implement the least-privileged security principle.",
        "Impact": "",
        "Remediation": "Refer to the policy syntax document and create new policies if the audit results indicate that the required policies are missing.",
        "Recommendation": "",
        "Observation": "custom IAM policy that grants tenancy administrative access."
    },
    "1.2": {
        "Description": "There is a built-in OCI IAM policy enabling the Administrators group to perform any action within a tenancy. In the OCI IAM console, this policy reads:<br><br><pre>\nAllow group Administrators to manage all-resources in tenancy\n</pre><br><br>Administrators create more users, groups, and policies to provide appropriate access to other groups.<br><br>Administrators should not allow any-other-group full access to the tenancy by writing a policy like this:<br><br><pre>\nAllow group any-other-group to manage all-resources in tenancy\n</pre><br><br>The access should be narrowed down to ensure the least-privileged principle is applied.",
        "Rationale": "Permission to manage all resources in a tenancy should be limited to a small number of users in the 'Administrators' group for break-glass situations and to set up users/groups/policies when a tenancy is created.<br><br>No group other than 'Administrators' in a tenancy should need access to all resources in a tenancy, as this violates the enforcement of the least privilege principle.",
        "Impact": "",
        "Remediation": "Remove any policy statement that allows any group other than Administrators or any service access to manage all resources in the tenancy.",
        "Recommendation": "Evaluate if tenancy-wide administrative access is needed for the identified policy and update it to be more restrictive.",
        "Observation": "custom IAM policy that grants tenancy administrative access."
    },
    "1.3": {
        "Description": "Tenancy administrators can create more users, groups, and policies to provide other service administrators access to OCI resources.<br><br>For example, an IAM administrator will need to have access to manage\n resources like compartments, users, groups, dynamic-groups, policies, identity-providers, tenancy tag-namespaces, tag-definitions in the tenancy.<br><br>The policy that gives IAM-Administrators or any other group full access to 'groups' resources should not allow access to the tenancy 'Administrators' group.<br><br>The policy statements would look like:<br><br><pre>\nAllow group IAMAdmins to inspect users in tenancy\nAllow group IAMAdmins to use users in tenancy where target.group.name != 'Administrators'\nAllow group IAMAdmins to inspect groups in tenancy\nAllow group IAMAdmins to use groups in tenancy where target.group.name != 'Administrators'\n</pre><br><br><b>Note:</b> You must include separate statements for 'inspect' access, because the target.group.name variable is not used by the ListUsers and ListGroups operations",
        "Rationale": "These policy statements ensure that no other group can manage tenancy administrator users or the membership to the 'Administrators' group thereby gain or remove tenancy administrator access.",
        "Impact": "",
        "Remediation": "Verify the results to ensure that the policy statements that grant access to use or manage users or groups in the tenancy have a condition that excludes access to Administrators group or to users in the Administrators group.",
        "Recommendation": "Evaluate if tenancy-wide administrative access is needed for the identified policy and update it to be more restrictive.",
        "Observation": "custom IAM policy that grants tenancy administrative access."
    },
    "1.4": {
        "Description": "Password policies are used to enforce password complexity requirements. IAM password policies can be used to ensure password are at least a certain length and are composed of certain characters.<br><br>It is recommended the password policy require a minimum password length 14 characters and contain 1 non-alphabetic\ncharacter (Number or 'Special Character').",
        "Rationale": "In keeping with the overall goal of having users create a password that is not overly weak, an eight-character minimum password length is recommended for an MFA account, and 14 characters for a password only account. In addition, maximum password length should be made as long as possible based on system/software capabilities and not restricted by policy.<br><br>In general, it is true that longer passwords are better (harder to crack), but it is also true that forced password length requirements can cause user behavior that is predictable and undesirable. For example, requiring users to have a minimum 16-character password may cause them to choose repeating patterns like fourfourfourfour or passwordpassword that meet the requirement but aren't hard to guess. Additionally, length requirements increase the chances that users will adopt other insecure practices, like writing them down, re-using them or storing them unencrypted in their documents. <br><br>Password composition requirements are a poor defense against guessing attacks. Forcing users to choose some combination of upper-case, lower-case, numbers, and special characters has a negative impact. It places an extra burden on users and many\nwill use predictable patterns (for example, a capital letter in the first position, followed by lowercase letters, then one or two numbers, and a “special character” at the end). Attackers know this, so dictionary attacks will often contain these common patterns and use the most common substitutions like, $ for s, @ for a, 1 for l, 0 for o.<br><br>Passwords that are too complex in nature make it harder for users to remember, leading to bad practices. In addition, composition requirements provide no defense against common attack types such as social engineering or insecure storage of passwords.",
        "Impact": "",
        "Remediation": "Update the password policy such as minimum length to 14, password must contain expected special characters and numeric characters.",
        "Recommendation": "It is recommended the password policy require a minimum password length 14 characters and contain 1 non-alphabetic character (Number or 'Special Character').",
        "Observation": "password policy/policies that do not enforce sufficient password complexity requirements."
    },
    "1.5": {
        "Description": "IAM password policies can require passwords to be rotated or expired after a given number of days. It is recommended that the password policy expire passwords after 365 and are changed immediately based on events.",
        "Rationale": "Excessive password expiration requirements do more harm than good, because these requirements make users select predictable passwords, composed of sequential words and numbers that are closely related to each other. In these cases, the next password can be predicted based on the previous one (incrementing a number used in the password for example). Also, password expiration requirements offer no containment benefits because attackers will often use credentials as soon as they compromise them. Instead, immediate password changes should be based on key events including, but not limited to:<br><br>1. Indication of compromise<br>2. Change of user roles<br>3. When a user leaves the organization.<br><br>Not only does changing passwords every few weeks or months frustrate the user, it's been suggested that it does more harm than good, because it could lead to bad practices by the user such as adding a character to the end of their existing password.<br><br>In addition, we also recommend a yearly password change. This is primarily because for all their good intentions users will share credentials across accounts. Therefore, even if a breach is publicly identified, the user may not see this notification, or forget they have an account on that site. This could leave a shared credential vulnerable indefinitely. Having an organizational policy of a 1-year (annual) password expiration is a reasonable compromise to mitigate this with minimal user burden.",
        "Impact": "",
        "Remediation": "Update the password policy by setting number of days configured in Expires after to 365.",
        "Recommendation": "Evaluate password rotation policies are inline with your organizational standard.",
        "Observation": "password policy/policies that do require rotation."
    },
    "1.6": {
        "Description": "IAM password policies can prevent the reuse of a given password by the same user. It is recommended the password policy prevent the reuse of passwords.",
        "Rationale": "Enforcing password history ensures that passwords are not reused in for a certain period of time by the same user. If a user is not allowed to use last 24 passwords, that window of time is greater. This helps maintain the effectiveness of password security.",
        "Impact": "",
        "Remediation": "Update the number of remembered passwords in previous passwords remembered setting to 24 in the password policy.",
        "Recommendation": "Evaluate password reuse policies are inline with your organizational standard.",
        "Observation": "password policy/policies that do prevent reuse."
    },
    "1.7": {
        "Description": "Multi-factor authentication is a method of authentication that requires the use of more than one factor to verify a user's identity.<br><br>With MFA enabled in the IAM service, when a user signs in to Oracle Cloud Infrastructure, they are prompted for their user name and password, which is the first factor (something that they know). The user is then prompted to provide a second verification code from a registered MFA device, which is the second factor (something that they have). The two factors work together, requiring an extra layer of security to verify the user's identity and complete the sign-in process.<br><br>OCI IAM supports two-factor authentication using a password (first factor) and a device that can generate a time-based one-time password (TOTP) (second factor).<br><br>See [OCI documentation](https://docs.cloud.oracle.com/en-us/iaas/Content/Identity/Tasks/usingmfa.htm) for more details.",
        "Rationale": "Multi factor authentication adds an extra layer of security during the login process and makes it harder for unauthorized users to gain access to OCI resources.",
        "Impact": "",
        "Remediation": "Each user must enable MFA for themselves using a device they will have access to every time they sign in. An administrator cannot enable MFA for another user but can enforce MFA by identifying the list of non-complaint users, notifying them or disabling access by resetting password for non-complaint accounts.",
        "Recommendation": "Evaluate if local users are required. For Break Glass accounts ensure MFA is in place.",
        "Observation": "users with Password access but not MFA."
    },
    "1.8": {
        "Description": "API keys are used by administrators, developers, services and scripts for accessing OCI APIs directly or via SDKs/OCI CLI to search, create, update or delete OCI resources.<br><br>The API key is an RSA key pair. The private key is used for signing the API requests and the public key is associated with a local or synchronized user's profile.",
        "Rationale": "It is important to secure and rotate an API key every 90 days or less as it provides the same level of access that a user it is associated with has.<br><br>In addition to a security engineering best practice, this is also a compliance requirement. For example, PCI-DSS Section 3.6.4 states, \"Verify that key-management procedures include a defined cryptoperiod for each key type in use and define a process for key changes at the end of the defined crypto period(s).\"",
        "Impact": "",
        "Remediation": "Delete any API Keys with a date of 90 days or older under the Created column of the API Key table.",
        "Recommendation": "Evaluate if APIs Keys are still used/required and rotate API Keys It is important to secure and rotate an API key every 90 days or less as it provides the same level of access that a user it is associated with has.",
        "Observation": "user(s) with APIs that have not been rotated with 90 days."
    },
    "1.9": {
        "Description": "Object Storage provides an API to enable interoperability with Amazon S3. To use this Amazon S3 Compatibility API, you need to generate the signing key required to authenticate with Amazon S3.<br><br>This special signing key is an Access Key/Secret Key pair. Oracle generates the Customer Secret key to pair with the Access Key.",
        "Rationale": "It is important to secure and rotate an customer secret key every 90 days or less as it provides the same level of object storage access that a user is associated with has.",
        "Impact": "",
        "Remediation": "Delete any Access Keys with a date of 90 days or older under the Created column of the Customer Secret Keys.",
        "Recommendation": "Evaluate if Customer Secret Keys are still used/required and rotate the Keys accordingly.",
        "Observation": "users with Customer Secret Keys that have not been rotated with 90 days."
    },
    "1.10": {
        "Description": "Auth tokens are authentication tokens generated by Oracle. You use auth tokens to authenticate with APIs that do not support the Oracle Cloud Infrastructure signature-based authentication. If the service requires an auth token, the service-specific documentation instructs you to generate one and how to use it.",
        "Rationale": "It is important to secure and rotate an auth token every 90 days or less as it provides the same level of access to APIs that do not support the OCI signature-based authentication as the user associated to it.",
        "Impact": "",
        "Remediation": "Delete any auth token with a date of 90 days or older under the Created column of the Auth Tokens.",
        "Recommendation": "Evaluate if Auth Tokens are still used/required and rotate Auth tokens.",
        "Observation": "user(s) with auth tokens that have not been rotated in 90 days."
    },
    "1.11": {
        "Description": "Users can create and manage their database password in their IAM user profile and use that password to authenticate to databases in their tenancy. An IAM database password is a different password than an OCI Console password. Setting an IAM database password allows an authorized IAM user to sign in to one or more Autonomous Databases in their tenancy. An IAM database password is a different password than an OCI Console password. Setting an IAM database password allows an authorized IAM user to sign in to one or more Autonomous Databases in their tenancy.",
        "Rationale": "It is important to secure and rotate an IAM Database password 90 days or less as it provides the same access the user would have a using a local database user.",
        "Impact": "",
        "Remediation": "Delete any database password with a date of 90 days or older under the Created column of the Database Password.",
        "Recommendation": "Evaluate if database password are still used/required and rotate database passwords.",
        "Observation": "user(s) with Database passwords that have not been rotated in 90 days."
    },
    "1.12": {
        "Description": "Tenancy administrator users have full access to the organization's OCI tenancy. API keys associated with user accounts are used for invoking the OCI APIs via custom programs or clients like CLI/SDKs. The clients are typically used for performing day-to-day operations and should never require full tenancy access. Service-level administrative users with API keys should be used instead.",
        "Rationale": "For performing day-to-day operations tenancy administrator access is not needed.\nService-level administrative users with API keys should be used to apply privileged security principle.",
        "Impact": "",
        "Remediation": "For each tenancy administrator user who has an API key,select API Keys from the menu and delete any associated keys from the API Keys table.",
        "Recommendation": "Evaluate if a user with API Keys requires Administrator access and use a least privilege approach.",
        "Observation": "users with Administrator access and API Keys."
    },
    "1.13": {
        "Description": "All OCI IAM local user accounts have an email address field associated with the account. It is recommended to specify an email address that is valid and current.<br><br>If you have an email address in your user profile, you can use the Forgot Password link on the sign on page to have a temporary password sent to you.",
        "Rationale": "Having a valid and current email address associated with an OCI IAM local user account allows you to tie the account to identity in your organization. It also allows that user to reset their password if it is forgotten or lost.",
        "Impact": "",
        "Remediation": "Update the current email address in the email text box on exch non compliant user.",
        "Recommendation": "Add emails to users to allow them to use the 'Forgot Password' feature and uniquely identify the user. For service accounts it could be a mail alias.",
        "Observation": "user(s) without an email."
    },
    "1.14": {
        "Description": "OCI instances, OCI database and OCI functions can access other OCI resources either via an OCI API key associated to a user or by being including in a Dynamic Group that has an IAM policy granting it the required access. Access to OCI Resources refers to making API calls to another OCI resource like Object Storage, OCI Vaults, etc.",
        "Rationale": "Dynamic Groups reduces the risks related to hard coded credentials. Hard coded API keys can be shared and require rotation which can open them up to being compromised. Compromised credentials could allow access to OCI services outside of the expected radius.",
        "Impact": "For an OCI instance that contains embedded credential audit the scripts and environment variables to ensure that none of them contain OCI API Keys or credentials.",
        "Remediation": "Create Dynamic group and Enter Matching Rules to that includes the instances accessing your OCI resources. Refer:\"https://docs.oracle.com/en-us/iaas/Content/Identity/Tasks/managingdynamicgroups.htm\".",
        "Recommendation": "Evaluate how your instances, functions, and autonomous database interact with other OCI services.",
        "Observation": "Dynamic Groups reduces the risks related to hard coded credentials. Hard coded API keys can be shared and require rotation which can open them up to being compromised. Compromised credentials could allow access to OCI services outside of the expected radius."
    },
    "1.15": {
        "Description": "To apply the separation of duties security principle, one can restrict service-level administrators from being able to delete resources they are managing. It means service-level administrators can only manage resources of a specific service but not delete resources for that specific service.<br><br>Example policies for global/tenant level for block volume service-administrators:\n<pre>\nAllow group VolumeUsers to manage volumes in tenancy where request.permission!='VOLUME_DELETE'\nAllow group VolumeUsers to manage volume-backups in tenancy where request.permission!='VOLUME_BACKUP_DELETE'\n</pre><br>Example policies for global/tenant level for file storage system service-administrators:<br><pre>\nAllow group FileUsers to manage file-systems in tenancy where request.permission!='FILE_SYSTEM_DELETE'\nAllow group FileUsers to manage mount-targets in tenancy where request.permission!='MOUNT_TARGET_DELETE'\nAllow group FileUsers to manage export-sets in tenancy where request.permission!='EXPORT_SET_DELETE'\n</pre><br><br>Example policies for global/tenant level for object storage system service-administrators:<br><pre>\nAllow group BucketUsers to manage objects in tenancy where request.permission!='OBJECT_DELETE'\nAllow group BucketUsers to manage buckets in tenancy where request.permission!='BUCKET_DELETE'\n</pre>",
        "Rationale": "Creating service-level administrators without the ability to delete the resource they are managing helps in tightly controlling access to Oracle Cloud Infrastructure (OCI) services by implementing the separation of duties security principle.", "Impact": "",
        "Remediation": "Add the appropriate where condition to any policy statement that allows the storage service-level to manage the storage service.",
        "Recommendation": "To apply a separation of duties security principle, it is recommended to restrict service-level administrators from being able to delete resources they are managing.",
        "Observation": "IAM Policies that give service administrator the ability to delete service resources."
    },
    "2.1": {
        "Description": "Security lists provide stateful or stateless filtering of ingress/egress network traffic to OCI resources on a subnet level. It is recommended that no security group allows unrestricted ingress access to port 22.",
        "Rationale": "Removing unfettered connectivity to remote console services, such as Secure Shell (SSH), reduces a server's exposure to risk.",
        "Impact": "For updating an existing environment, care should be taken to ensure that administrators currently relying on an existing ingress from 0.0.0.0/0 have access to ports 22 and/or 3389 through another network security group or security list.",
        "Remediation": "For each security list in the returned results, click the security list name. Either edit the ingress rule to be more restrictive, delete the ingress rule or click on the VCN and terminate the security list as appropriate.",
        "Recommendation": "Review the security lists. If they are not used(attached to a subnet) they should be deleted if possible or empty. For attached security lists it is recommended to restrict the CIDR block to only allow access to Port 22 from known networks.",
        "Observation": "Security lists that allow internet access to port 22. (Note this does not necessarily mean external traffic can reach a compute instance)."
    },
    "2.2": {
        "Description": "Security lists provide stateful or stateless filtering of ingress/egress network traffic to OCI resources on a subnet level. It is recommended that no security group allows unrestricted ingress access to port 3389.",
        "Rationale": "Removing unfettered connectivity to remote console services, such as Remote Desktop Protocol (RDP), reduces a server's exposure to risk.",
        "Impact": "For updating an existing environment, care should be taken to ensure that administrators currently relying on an existing ingress from 0.0.0.0/0 have access to ports 22 and/or 3389 through another network security group or security list.",
        "Remediation": "For each security list in the returned results, click the security list name. Either edit the ingress rule to be more restrictive, delete the ingress rule or click on the VCN and terminate the security list as appropriate.",
        "Recommendation": "Review the security lists. If they are not used(attached to a subnet) they should be deleted if possible or empty. For attached security lists it is recommended to restrict the CIDR block to only allow access to Port 3389 from known networks.",
        "Observation": "Security lists that allow internet access to port 3389. (Note this does not necessarily mean external traffic can reach a compute instance)."
    },
    "2.3": {
        "Description": "Network security groups provide stateful filtering of ingress/egress network traffic to OCI resources. It is recommended that no security group allows unrestricted ingress access to port 22.",
        "Rationale": "Removing unfettered connectivity to remote console services, such as Secure Shell (SSH), reduces a server's exposure to risk.",
        "Impact": "For updating an existing environment, care should be taken to ensure that administrators currently relying on an existing ingress from 0.0.0.0/0 have access to ports 22 and/or 3389 through another network security group or security list.",
        "Remediation": "Using the details returned from the audit procedure either Remove the security rules or Update the security rules.",
        "Recommendation": "Review the network security groups. If they are not used(attached to a subnet) they should be deleted if possible or empty. For attached security lists it is recommended to restrict the CIDR block to only allow access to Port 3389 from known networks.",
        "Observation": "Network security groups that allow internet access to port 22. (Note this does not necessarily mean external traffic can reach a compute instance)."
    },
    "2.4": {
        "Description": "Network security groups provide stateful filtering of ingress/egress network traffic to OCI resources. It is recommended that no security group allows unrestricted ingress access to port 3389.",
        "Rationale": "Removing unfettered connectivity to remote console services, such as Remote Desktop Protocol (RDP), reduces a server's exposure to risk.",
        "Impact": "For updating an existing environment, care should be taken to ensure that administrators currently relying on an existing ingress from 0.0.0.0/0 have access to ports 22 and/or 3389 through another network security group or security list.",
        "Remediation": "Using the details returned from the audit procedure either Remove the security rules or Update the security rules.",
        "Recommendation": "Review the network security groups. If they are not used(attached to a subnet) they should be deleted if possible or empty. For attached network security groups it is recommended to restrict the CIDR block to only allow access to Port 3389 from known networks.",
        "Observation": "Network security groups that allow internet access to port 3389. (Note this does not necessarily mean external traffic can reach a compute instance)."
    },
    "2.5": {
        "Description": "A default security list is created when a Virtual Cloud Network (VCN) is created. Security lists provide stateful filtering of ingress and egress network traffic to OCI resources. It is recommended no security list allows unrestricted ingress access to Secure Shell (SSH) via port 22.",
        "Rationale": "Removing unfettered connectivity to remote console services, such as SSH on port 22, reduces a server's exposure to unauthorized access.",
        "Impact": "For updating an existing environment, care should be taken to ensure that administrators currently relying on an existing ingress from 0.0.0.0/0 have access to ports 22 and/or 3389 through another security group.",
        "Remediation": "Select Default Security List for <VCN Name> and Remove the Ingress Rule with Source 0.0.0.0/0, IP Protocol 22 and Destination Port Range 22.",
        "Recommendation": "Create specific custom security lists with workload specific rules and attach to subnets.",
        "Observation": "Default Security lists that allow more traffic then ICMP."
    },
    "2.6": {
        "Description": "Oracle Integration (OIC) is a complete, secure, but lightweight integration solution that enables you to connect your applications in the cloud. It simplifies connectivity between your applications and connects both your applications that live in the cloud and your applications that still live on premises. Oracle Integration provides secure, enterprise-grade connectivity regardless of the applications you are connecting or where they reside. OIC instances are created within an Oracle managed secure private network with each having a public endpoint. The capability to configure ingress filtering of network traffic to protect your OIC instances from unauthorized network access is included. It is recommended that network access to your OIC instances be restricted to your approved corporate IP Addresses or Virtual Cloud Networks (VCN)s.",
        "Rationale": "Restricting connectivity to OIC Instances reduces an OIC instance's exposure to risk.",
        "Impact": "When updating ingress filters for an existing environment, care should be taken to ensure that IP addresses and VCNs currently used by administrators, users, and services to access your OIC instances are included in the updated filters.",
        "Remediation": "For each OIC instance in the returned results, select the OIC Instance name,edit the Network Access to be more restrictive.",
        "Recommendation": "It is recommended that OIC Network Access is restricted to your corporate IP Addresses or VCNs for OIC Instances.",
        "Observation": "OIC Instances that allow unfiltered public ingress traffic (Authentication and authorization is still required)."
    },
    "2.7": {
        "Description": "Oracle Analytics Cloud (OAC) is a scalable and secure public cloud service that provides a full set of capabilities to explore and perform collaborative analytics for you, your workgroup, and your enterprise. OAC instances provide ingress filtering of network traffic or can be deployed with in an existing Virtual Cloud Network VCN. It is recommended that all new OAC instances be deployed within a VCN and that the Access Control Rules are restricted to your corporate IP Addresses or VCNs for existing OAC instances.",
        "Rationale": "Restricting connectivity to Oracle Analytics Cloud instances reduces an OAC instance's exposure to risk.",
        "Impact": "When updating ingress filters for an existing environment, care should be taken to ensure that IP addresses and VCNs currently used by administrators, users, and services to access your OAC instances are included in the updated filters. Also, these changes will temporarily bring the OAC instance offline.",
        "Remediation": "For each OAC instance in the returned results, select the OAC Instance name edit the Access Control Rules by clicking +Another Rule and add rules as required.",
        "Recommendation": "It is recommended that all new OAC instances be deployed within a VCN and that the Access Control Rules are restricted to your corporate IP Addresses or VCNs for existing OAC instances.",
        "Observation": "OAC Instances that allow unfiltered public ingress traffic (Authentication and authorization is still required)."
    },
    "2.8": {
        "Description": "Oracle Autonomous Database Shared (ADB-S) automates database tuning, security, backups, updates, and other routine management tasks traditionally performed by DBAs. ADB-S provide ingress filtering of network traffic or can be deployed within an existing Virtual Cloud Network (VCN). It is recommended that all new ADB-S databases be deployed within a VCN and that the Access Control Rules are restricted to your corporate IP Addresses or VCNs for existing ADB-S databases.",
        "Rationale": "Restricting connectivity to ADB-S Databases reduces an ADB-S database's exposure to risk.",
        "Impact": "When updating ingress filters for an existing environment, care should be taken to ensure that IP addresses and VCNs currently used by administrators, users, and services to access your ADB-S instances are included in the updated filters.",
        "Remediation": "For each ADB-S database in the returned results, select the ADB-S database name edit the Access Control Rules by clicking +Another Rule and add rules as required.",
        "Recommendation": "It is recommended that all new ADB-S databases be deployed within a VCN and that the Access Control Rules are restricted to your corporate IP Addresses or VCNs for existing ADB-S databases.",
        "Observation": "ADB-S Instances that allow unfiltered public ingress traffic (Authentication and authorization is still required)."
    },
    "3.1": {
        "Description": "Compute Instances that utilize Legacy MetaData service endpoints (IMDSv1) are susceptible to potential SSRF attacks. To bolster security measures, it is strongly advised to reconfigure Compute Instances to adopt Instance Metadata Service v2, aligning with the industry's best security practices.",
        "Rationale": "Enabling Instance Metadata Service v2 enhances security and grants precise control over metadata access. Transitioning from IMDSv1 reduces the risk of SSRF attacks, bolstering system protection.  IMDv1 poses security risks due to its inferior security measures and limited auditing capabilities. Transitioning to IMDv2 ensures a more secure environment with robust security features and improved monitoring capabilities.",
        "Impact": "If you disable IMDSv1 on an instance that does not support IMDSv2, you might not be able to connect to the instance when you launch it.",
        "Remediation": "For each instance select the instance name, under the Instance Details section, next to Instance Metadata Service, click Edit and for the Instance metadata service, select the Version 2 only option.",
        "Recommendation": "It is recommended that all OCI instances use Instance Metadata Service version 2 (IMDSv2).",
        "Observation": "Instances that allow Instance Metadata Service v1."
    },
    "3.2": {
        "Description": "Shielded Instances with Secure Boot enabled prevents unauthorized boot loaders and operating systems from booting. This prevent rootkits, bootkits, and unauthorized software from running before the operating system loads. Secure Boot verifies the digital signature of the system's boot software to check its authenticity. The digital signature ensures the operating system has not been tampered with and is from a trusted source. When the system boots and attempts to execute the software, it will first check the digital signature to ensure validity. If the digital signature is not valid, the system will not allow the software to run. Secure Boot is a feature of UEFI(Unified Extensible Firmware Interface) that only allows approved operating systems to boot up.",
        "Rationale": "A Threat Actor with access to the operating system may seek to alter boot components to persist malware or rootkits during system initialization. Secure Boot helps ensure that the system only runs authentic software by verifying the digital signature of all boot components.",
        "Impact": " To enable you have to terminate the instance and create a new one. Also, Shielded instances do not support live migration. During an infrastructure maintenance event, Oracle Cloud Infrastructure live migrates supported VM instances from the physical VM host that needs maintenance to a healthy VM host with minimal disruption to running instances. If you enable Secure Boot on an instance, the instance cannot be migrated, because the hardware TPM is not migratable. This may result in an outage because the TPM can't be migrate from a unhealthy host to healthy host.",
        "Remediation": "Terminate the old instance. Create a new instance and ensure on Secure Boot is toggled on under the Security section.",
        "Recommendation": "",
        "Observation": "Instances that don't enable Secure Boot."
    },
    "3.3": {
        "Description": "The Block Volume service provides the option to enable in-transit encryption for paravirtualized volume attachments on virtual machine (VM) instances.",
        "Rationale": "All the data moving between the instance and the block volume is transferred over an internal and highly secure network. If you have specific compliance requirements related to the encryption of the data while it is moving between the instance and the block volume, you should enable the in-transit encryption option.",
        "Impact": "In-transit encryption for boot and block volumes is only available for virtual machine (VM) instances launched from platform images, along with bare metal instances that use the following shapes: BM.Standard.E3.128, BM.Standard.E4.128, BM.DenseIO.E4.128. It is not supported on other bare metal instances.",
        "Remediation": "Terminate the old instance. Create a new instance and ensure Use in-transit encryption is toggled on under the Boot volume section.",
        "Recommendation": "",
        "Observation": "Instances that don't enable in-transit encryption."
    },
    "4.1": {
        "Description": "Using default tags is a way to ensure all resources that support tags are tagged during creation. Tags can be based on static values or based on computed values. It is recommended to setup default tags early on to ensure all created resources will get tagged.\nTags are scoped to Compartments and are inherited by Child Compartments. The recommendation is to create default tags like “CreatedBy” at the Root Compartment level to ensure all resources get tagged.\nWhen using Tags it is important to ensure that Tag Namespaces are protected by IAM Policies otherwise this will allow users to change tags or tag values.\nDepending on the age of the OCI Tenancy there may already be Tag defaults setup at the Root Level and no need for further action to implement this action.",
        "Rationale": "In the case of an incident having default tags like “CreatedBy” applied will provide info on who created the resource without having to search the Audit logs.",
        "Impact": "There is no performance impact when enabling the above described features",
        "Remediation": "Update the root compartments tag default link.In the Tag Defaults table verify that there is a Tag with a value of \"${iam.principal.names}\" and a Tag Key Status of Active. Also create a Tag key definition by providing a Tag Key, Description and selecting 'Static Value' for Tag Value Type.",
        "Recommendation": "",
        "Observation": "default tags are used on resources."
    },
    "4.2": {
        "Description": "Notifications provide a multi-channel messaging service that allow users and applications to be notified of events of interest occurring within OCI. Messages can be sent via eMail, HTTPs, PagerDuty, Slack or the OCI Function service. Some channels, such as eMail require confirmation of the subscription before it becomes active.",
        "Rationale": "Creating one or more notification topics allow administrators to be notified of relevant changes made to OCI infrastructure.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Create a Topic in the notifications service under the appropriate compartment and add the subscriptions with current email address and correct protocol.",
        "Recommendation": "",
        "Observation": "notification topic and subscription for receiving monitoring alerts are configured."
    },
    "4.3": {
        "Description": "It is recommended to setup an Event Rule and Notification that gets triggered when Identity Providers are created, updated or deleted. Event Rules are compartment scoped and will detect events in child compartments. It is recommended to create the Event rule at the root compartment level.",
        "Rationale": "OCI Identity Providers allow management of User ID / passwords in external systems and use of those credentials to access OCI resources. Identity Providers allow users to single sign-on to OCI console and have other OCI credentials like API Keys.\nMonitoring and alerting on changes to Identity Providers will help in identifying changes to the security posture.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Create a Rule Condition in the Events services by selecting Identity in the Service Name Drop-down and selecting Identity Provider – Create, Identity Provider - Delete and Identity Provider – Update. In the Actions section select Notifications as Action Type and selct the compartment and topic to be used.",
        "Recommendation": "",
        "Observation": "notifications have been configured for Identity Provider changes."
    },
    "4.4": {
        "Description": "It is recommended to setup an Event Rule and Notification that gets triggered when Identity Provider Group Mappings are created, updated or deleted. Event Rules are compartment scoped and will detect events in child compartments. It is recommended to create the Event rule at the root compartment level",
        "Rationale": "IAM Policies govern access to all resources within an OCI Tenancy. IAM Policies use OCI Groups for assigning the privileges. Identity Provider Groups could be mapped to OCI Groups to assign privileges to federated users in OCI. Monitoring and alerting on changes to Identity Provider Group mappings will help in identifying changes to the security posture.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Find and click the Rule that handles Idp Group Mapping Changes. Click the Edit Rule button and verify that the RuleConditions section contains a condition for the Service Identity and Event Types: Idp Group Mapping – Create, Idp Group Mapping – Delete, and Idp Group Mapping – Update and confirm Action Type contains: Notifications and that a valid Topic is referenced.",
        "Recommendation": "",
        "Observation": "notifications have been configured for Identity Provider Group Mapping changes."
    },
    "4.5": {
        "Description": "It is recommended to setup an Event Rule and Notification that gets triggered when IAM Groups are created, updated or deleted. Event Rules are compartment scoped and will detect events in child compartments, it is recommended to create the Event rule at the root compartment level.",
        "Rationale": "IAM Groups control access to all resources within an OCI Tenancy.\n Monitoring and alerting on changes to IAM Groups will help in identifying changes to satisfy least privilege principle.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Create a Rule Condition by selecting Identity in the Service Name Drop-down and selecting Group – Create, Group – Delete and Group – Update. In the Actions section select Notifications as Action Type and selct the compartment and topic to be used.",
        "Recommendation": "",
        "Observation": "notifications have been configured for IAM Group changes."
    },
    "4.6": {
        "Description": "It is recommended to setup an Event Rule and Notification that gets triggered when IAM Policies are created, updated or deleted. Event Rules are compartment scoped and will detect events in child compartments, it is recommended to create the Event rule at the root compartment level.",
        "Rationale": "IAM Policies govern access to all resources within an OCI Tenancy.\n Monitoring and alerting on changes to IAM policies will help in identifying changes to the security posture.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Create a Rule Condition by selecting Identity in the Service Name Drop-down and selecting Policy – Change Compartment, Policy – Create, Policy - Delete and Policy – Update. In the Actions section select Notifications as Action Type and selct the compartment and topic to be used.",
        "Recommendation": "",
        "Observation": "notifications have been configured for IAM Policy changes."
    },
    "4.7": {
        "Description": "It is recommended to setup an Event Rule and Notification that gets triggered when IAM Users are created, updated, deleted, capabilities updated, or state updated. Event Rules are compartment scoped and will detect events in child compartments, it is recommended to create the Event rule at the root compartment level.",
        "Rationale": "Users use or manage Oracle Cloud Infrastructure resources.\n Monitoring and alerting on changes to Users will help in identifying changes to the security posture.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Edit Rule that handles IAM User Changes and verify that the Rule Conditions section contains a condition for the Service Identity and Event Types: User – Create, User – Delete, User – Update, User Capabilities – Update, User State – Update.",
        "Recommendation": "",
        "Observation": "notifications have been configured for user changes."
    },
    "4.8": {
        "Description": "It is recommended to setup an Event Rule and Notification that gets triggered when Virtual Cloud Networks are created, updated or deleted. Event Rules are compartment scoped and will detect events in child compartments, it is recommended to create the Event rule at the root compartment level.",
        "Rationale": "Virtual Cloud Networks (VCNs) closely resembles a traditional network.\n Monitoring and alerting on changes to VCNs will help in identifying changes to the security posture.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Edit Rule that handles VCN Changes and verify that the RuleConditions section contains a condition for the Service Networking and Event Types: VCN – Create, VCN - Delete, and VCN – Update.",
        "Recommendation": "",
        "Observation": "notifications have been configured for VCN changes."
    },
    "4.9": {
        "Description": "It is recommended to setup an Event Rule and Notification that gets triggered when route tables are created, updated or deleted. Event Rules are compartment scoped and will detect events in child compartments, it is recommended to create the Event rule at the root compartment level.",
        "Rationale": "Route tables control traffic flowing to or from Virtual Cloud Networks and Subnets.\n Monitoring and alerting on changes to route tables will help in identifying changes these traffic flows.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Edit Rule that handles Route Table Changes and verify that the RuleConditions section contains a condition for the Service Networking and Event Types: Route Table – Change Compartment, Route Table – Create, Route Table - Delete, and Route Table – Update.",
        "Recommendation": "",
        "Observation": "notifications have been configured for changes to route tables."
    },
    "4.10": {
        "Description": "It is recommended to setup an Event Rule and Notification that gets triggered when security lists are created, updated or deleted. Event Rules are compartment scoped and will detect events in child compartments, it is recommended to create the Event rule at the root compartment level.",
        "Rationale": "Security Lists control traffic flowing into and out of Subnets within a Virtual Cloud Network.\n Monitoring and alerting on changes to Security Lists will help in identifying changes to these security controls.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Edit Rule that handles Security List Changes and verify that the RuleConditions section contains a condition for the Service Networking and Event Types: Security List – Change Compartment, Security List – Create, Security List - Delete, and Security List – Update.",
        "Recommendation": "",
        "Observation": "notifications have been configured for security list changes."
    },
    "4.11": {
        "Description": "It is recommended to setup an Event Rule and Notification that gets triggered when network security groups are created, updated or deleted. Event Rules are compartment scoped and will detect events in child compartments, it is recommended to create the Event rule at the root compartment level.",
        "Rationale": "Network Security Groups control traffic flowing between Virtual Network Cards attached to Compute instances.\n Monitoring and alerting on changes to Network Security Groups will help in identifying changes these security controls.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Edit Rule that handles Network Security Group changes and verify that the RuleConditions section contains a condition for the Service Networking and Event Types: Network Security Group – Change Compartment, Network Security Group – Create, Network Security Group - Delete, and Network Security Group – Update.",
        "Recommendation": "",
        "Observation": "notifications have been configured for changes on Network Service Groups."
    },
    "4.12": {
        "Description": "It is recommended to setup an Event Rule and Notification that gets triggered when Network Gateways are created, updated, deleted, attached, detached, or moved. This recommendation includes Internet Gateways, Dynamic Routing Gateways, Service Gateways, Local Peering Gateways, and NAT Gateways. Event Rules are compartment scoped and will detect events in child compartments, it is recommended to create the Event rule at the root compartment level.",
        "Rationale": "Network Gateways act as routers between VCNs and the Internet, Oracle Services Networks, other VCNS, and on-premise networks.\n Monitoring and alerting on changes to Network Gateways will help in identifying changes to the security posture.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Edit Rule that handles Network Gateways Changes and verify that the RuleConditions section contains a condition for the Service Networking and Event Types: DRG – Create, DRG - Delete, DRG - Update, DRG Attachment – Create, DRG Attachment – Delete, DRG Attachment - Update, Internet Gateway – Create, Internet Gateway – Delete, Internet Gateway - Update, Internet Gateway – Change Compartment, Local Peering Gateway – Create, Local Peering Gateway – Delete End, Local Peering Gateway - Update, Local Peering Gateway – Change Compartment, NAT Gateway – Create, NAT Gateway – Delete, NAT Gateway - Update, NAT Gateway – Change Compartment,Compartment, Service Gateway – Create, Service Gateway – Delete Begin, Service Gateway – Delete End, Service Gateway – Update, Service Gateway – Attach Service, Service Gateway – Detach Service, Service Gateway – Change Compartment.",
        "Recommendation": "",
        "Observation": "notifications have been configured for changes on network gateways."
    },
    "4.13": {
        "Description": "VCN flow logs record details about traffic that has been accepted or rejected based on the security list rule.",
        "Rationale": "Enabling VCN flow logs enables you to monitor traffic flowing within your virtual network and can be used to detect anomalous traffic.",
        "Impact": "Enabling VCN flow logs will not affect the performance of your virtual network but it will generate additional use of object storage that should be controlled via object lifecycle management.<br><br>By default, VCN flow logs are stored for 30 days in object storage. Users can specify a longer retention period.",
        "Remediation": "Enable Flow Logs (all records) on Virtual Cloud Networks (subnets) under the relevant resource compartment. Before hand create Log group if not exist in the Log services.",
        "Recommendation": "",
        "Observation": "VCNs have no flow logging configured."
    },
    "4.14": {
        "Description": "Cloud Guard detects misconfigured resources and insecure activity within a tenancy and provides security administrators with the visibility to resolve these issues. Upon detection, Cloud Guard can suggest, assist, or take corrective actions to mitigate these issues. Cloud Guard should be enabled in the root compartment of your tenancy with the default configuration, activity detectors and responders.",
        "Rationale": "Cloud Guard provides an automated means to monitor a tenancy for resources that are configured in an insecure manner as well as risky network activity from these resources.",
        "Impact": "There is no performance impact when enabling the above described features, but additional IAM policies will be required.",
        "Remediation": "Enable the cloud guard by selecting the services in the menu and provide appropriate reporting region and other configurations.",
        "Recommendation": "",
        "Observation": "Cloud Guard has not been configured in the root compartment of the tenancy."
    },
    "4.15" : {
        "Description": "Cloud Guard detects misconfigured resources and insecure activity within a tenancy and provides security administrators with the visibility to resolve these issues. Upon detection, Cloud Guard generates a Problem. It is recommended to setup an Event Rule and Notification that gets triggered when Oracle Cloud Guard Problems are created, dismissed or remediated. Event Rules are compartment scoped and will detect events in child compartments. It is recommended to create the Event rule at the root compartment level.",
        "Rationale": "Cloud Guard provides an automated means to monitor a tenancy for resources that are configured in an insecure manner as well as risky network activity from these resources. Monitoring and alerting on Problems detected by Cloud Guard will help in identifying changes to the security posture.",
        "Impact": "There is no performance impact when enabling the above described features but depending on the amount of notifications sent per month there may be a cost associated.",
        "Remediation": "Create a Rule Condition by selecting Cloud Guard in the Service Name Drop-down and selecting Detected – Problem, Remediated – Problem and Dismissed - Problem. In the Actions section select Notifications as Action Type and selct the compartment and topic to be used.",
        "Recommendation": "",
        "Observation": "notifications have been configured for Cloud Guard Problems."
    },
    "4.16": {
        "Description": "Oracle Cloud Infrastructure Vault securely stores master encryption keys that protect your encrypted data. You can use the Vault service to rotate keys to generate new cryptographic material. Periodically rotating keys limits the amount of data encrypted by one key version.",
        "Rationale": "Rotating keys annually limits the data encrypted under one key version. Key rotation thereby reduces the risk in case a key is ever compromised.",
        "Impact": "",
        "Remediation": "Select the security service and select vault. Ensure the date of each Master Encryption Key under the Created column of the Master Encryption key is no more than 365 days old.",
        "Recommendation": "",
        "Observation": "customer-managed keys are older than one year."
    },
    "4.17": {
        "Description": "Object Storage write logs will log all write requests made to objects in a bucket.",
        "Rationale": "Enabling an Object Storage write log, the 'requestAction' property would contain values of 'PUT', 'POST', or 'DELETE'. This will provide you more visibility into changes to objects in your buckets.",
        "Impact": "There is no performance impact when enabling the above described features, but will generate additional use of object storage that should be controlled via object lifecycle management.<br><br>By default, Object Storage logs are stored for 30 days in object storage. Users can specify a longer retention period.",
        "Remediation": "To the relevant bucket enable log by providing Write Access Events from the Log Category. Beforehand create log group if required.",
        "Recommendation": "",
        "Observation": "object stores have no write level logging enabled."
    },
    "5.1.1": {
        "Description": "A bucket is a logical container for storing objects. It is associated with a single compartment that has policies that determine what action a user can perform on a bucket and on all the objects in the bucket. It is recommended that no bucket be publicly accessible.",
        "Rationale": "Removing unfettered reading of objects in a bucket reduces an organization's exposure to data loss.",
        "Impact": "For updating an existing bucket, care should be taken to ensure objects in the bucket can be accessed through either IAM policies or pre-authenticated requests.",
        "Remediation": "Edit the visibility into 'private' for each Bucket.",
        "Recommendation": "",
        "Observation": "object storage buckets are publicly visible."
    },
    "5.1.2": {
        "Description": "Oracle Object Storage buckets support encryption with a Customer Managed Key (CMK). By default, Object Storage buckets are encrypted with an Oracle managed key.",
        "Rationale": "Encryption of Object Storage buckets with a Customer Managed Key (CMK) provides an additional level of security on your data by allowing you to manage your own encryption key lifecycle management for the bucket.",
        "Impact": "Encrypting with a Customer Managed Keys requires a Vault and a Customer Master Key. In addition, you must authorize Object Storage service to use keys on your behalf.<br><br>Required Policy:\n<pre>\nAllow service objectstorage-&lt;region_name>, to use keys in compartment &ltcompartment-id> where target.key.id = '&lt;key_OCID>'<br><br></pre>",
        "Remediation": "Assign Master encryption key to Encryption key in every Object storage under Bucket name by clicking assign and select vault.",
        "Recommendation": "",
        "Observation": "object store buckets do not use Customer-Managed Keys (CMK)."
    },
    "5.1.3": {
        "Description": "A bucket is a logical container for storing objects. Object versioning is enabled at the bucket level and is disabled by default upon creation. Versioning directs Object Storage to automatically create an object version each time a new object is uploaded, an existing object is overwritten, or when an object is deleted. You can enable object versioning at bucket creation time or later.",
        "Rationale": "Versioning object storage buckets provides for additional integrity of your data. Management of data integrity is critical to protecting and accessing protected data. Some customers want to identify object storage buckets without versioning in order to apply their own data lifecycle protection and management policy.",
        "Impact": "",
        "Remediation": "Enable Versioning by clicking on every bucket by editing the bucket configuration.",
        "Recommendation": "",
        "Observation": "object store buckets have no versioning enabled."
    },
    "5.2.1": {
        "Description": "Oracle Cloud Infrastructure Block Volume service lets you dynamically provision and manage block storage volumes. By default, the Oracle service manages the keys that encrypt this block volume. Block Volumes can also be encrypted using a customer managed key.",
        "Rationale": "Encryption of block volumes provides an additional level of security for your data. Management of encryption keys is critical to protecting and accessing protected data. Customers should identify block volumes encrypted with Oracle service managed keys in order to determine if they want to manage the keys for certain volumes and then apply their own key lifecycle management to the selected block volumes.",
        "Impact": "Encrypting with a Customer Managed Keys requires a Vault and a Customer Master Key. In addition, you must authorize the Block Volume service to use the keys you create.\nRequired IAM Policy:\n<pre>\nAllow service blockstorage to use keys in compartment &ltcompartment-id> where target.key.id = '&lt;key_OCID>'\n</pre>",
        "Remediation": "For each block volumes from the result, assign the encryption key by Selecting the Vault Compartment and Vault, select the Master Encryption Key Compartment and Master Encryption key, click Assign.",
        "Recommendation": "",
        "Observation": "block volumes are not encrypted with a Customer-Managed Key."
    },
    "5.2.2": {
        "Description": "When you launch a virtual machine (VM) or bare metal instance based on a platform image or custom image, a new boot volume for the instance is created in the same compartment. That boot volume is associated with that instance until you terminate the instance. By default, the Oracle service manages the keys that encrypt this boot volume. Boot Volumes can also be encrypted using a customer managed key.",
        "Rationale": "Encryption of boot volumes provides an additional level of security for your data. Management of encryption keys is critical to protecting and accessing protected data. Customers should identify boot volumes encrypted with Oracle service managed keys in order to determine if they want to manage the keys for certain boot volumes and then apply their own key lifecycle management to the selected boot volumes.",
        "Impact": "Encrypting with a Customer Managed Keys requires a Vault and a Customer Master Key. In addition, you must authorize the Boot Volume service to use the keys you create.\nRequired IAM Policy:\n<pre>\nAllow service Bootstorage to use keys in compartment &ltcompartment-id> where target.key.id = '&lt;key_OCID>'\n</pre>",
        "Remediation": "For each boot volumes from the result, assign the encryption key by Selecting the Vault Compartment and Vault, select the Master Encryption Key Compartment and Master Encryption key, click Assign.",
        "Recommendation": "",
        "Observation": "boot volumes are not encrypted with a Customer-Managed Key."
    },
    "5.3.1": {
        "Description": "Oracle Cloud Infrastructure File Storage service (FSS) provides a durable, scalable, secure, enterprise-grade network file system. By default, the Oracle service manages the keys that encrypt FSS file systems. FSS file systems can also be encrypted using a customer managed key.",
        "Rationale": "Encryption of FSS systems provides an additional level of security for your data. Management of encryption keys is critical to protecting and accessing protected data. Customers should identify FSS file systems that are encrypted with Oracle service managed keys in order to determine if they want to manage the keys for certain FSS file systems and then apply their own key lifecycle management to the selected FSS file systems.",
        "Impact": "Encrypting with a Customer Managed Keys requires a Vault and a Customer Master Key. In addition, you must authorize the File Storage service to use the keys you create.\nRequired IAM Policy:\n<pre>\nAllow service FssOc1Prod to use keys in compartment &ltcompartment-id> where target.key.id = '&lt;key_OCID>'\n</pre>",
        "Remediation": "For each file storage system from the result, assign the encryption key by Selecting the Vault Compartment and Vault, select the Master Encryption Key Compartment and Master Encryption key, click Assign.",
        "Recommendation": "",
        "Observation": "file storage services (FSS) are not encrypted with a Customer-Managed Key."
    },
    "6.1": {
        "Description": "When you sign up for Oracle Cloud Infrastructure, Oracle creates your tenancy, which is the root compartment that holds all your cloud resources. You then create additional compartments within the tenancy (root compartment) and corresponding policies to control access to the resources in each compartment.<br><br>Compartments allow you to organize and control access to your cloud resources. A compartment is a collection of related resources (such as instances, databases, virtual cloud networks, block volumes) that can be accessed only by certain groups that have been given permission by an administrator.",
        "Rationale": "Compartments are a logical group that adds an extra layer of isolation, organization and authorization making it harder for unauthorized users to gain access to OCI resources.",
        "Impact": "Once the compartment is created an OCI IAM policy must be created to allow a group to resources in the compartment otherwise only group with tenancy access will have access.",
        "Remediation": "Create the new compartment under the root compartment.",
        "Recommendation": "",
        "Observation": "Only the root compartment is used in the tenancy."
    },
    "6.2": {
        "Description": "When you create a cloud resource such as an instance, block volume, or cloud network, you must specify to which compartment you want the resource to belong. Placing resources in the root compartment makes it difficult to organize and isolate those resources.",
        "Rationale": "Placing resources into a compartment will allow you to organize and have more granular access controls to your cloud resources.",
        "Impact": "Placing a resource in a compartment will impact how you write policies to manage access and organize that resource.",
        "Remediation": "For each item in the returned results, select Move Resource or More Actions then Move Resource and select compartment except root and choose new then move resources.",
        "Recommendation": "",
        "Observation": "resources are created in the root compartment."
    }
}



import json
import requests
import logging
import urllib3

from . import utils

# Disable warnings for insecure HTTPS requests (optional, for self-signed certs)
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# Get Auth Token
def get_token(args):
    url = args.url + "/dna/system/api/v1/auth/token"
    response = requests.post(url, auth=(args.user, args.password), verify=False)
    response.raise_for_status()
    token = response.json()["Token"]
    return token

# Get network devices
def get_devices(args, token):
    url = args.url + "dna/intent/api/v1/network-device"
    headers = {'X-Auth-Token': token}
    response = requests.get(url, headers=headers, verify=False)
    response.raise_for_status()
    return response.json().get('response', [])

# Main entry point
def get_inventory(args):
    assets = []
    logging.info("Retrieving network inventory from Cisco DNA Center") 
    token = get_token(args)
    devices = get_devices(args, token)

    for device in devices:
        asset = {}
        asset['owner'] = args.handle
        asset['id'] = device['managementIpAddress']
        asset['name'] = device['hostname']
        asset['type'] = 'Cisco'
        asset['tags'] = 'Cisco DNA'
        #asset['tags'] = ['Model:'+device['platformId'], 'Serial Number:'+device['serialNumber']]
        products = ['Cisco '+device['platformId']+' '+device['softwareVersion']]
        asset['products'] = products
    return assets



import sys
import platform
import os
import tempfile
import shutil
import uuid
import subprocess
import logging
import socket
import csv
import ipaddress
import getpass
import base64
import json
import pkg_resources
import importlib
import traceback
import warnings
with warnings.catch_warnings():
   warnings.simplefilter("ignore", category=Warning)
   from cryptography.hazmat.backends import default_backend
   from cryptography.hazmat.primitives import hashes
   from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
   from cryptography.fernet import Fernet
from . import utils
from . import plugin_processor
from .host_benchmark import host_benchmark as host_benchmark

def check_host_up(host):
    if not host['remote']:
        return True
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(2)
    try:
        result = sock.connect_ex((host['hostname'], 22))
        if result == 0:
            return True
        else:
            return False
    except socket.timeout:
        logging.error("Socket timeout")
        return False

def discover_openbsd(args, host):
    plist = []
    cmdarr = ["/usr/sbin/pkg_info -A"]
    logging.info("Retrieving product details")
    pkgout = utils.run_cmd_on_host(args, host, cmdarr)

    begin = False
    for l in pkgout.splitlines():
        lsplit = l.split()
        pkgline = lsplit[0]
        ldash = pkgline.rfind('-')
        pkg = pkgline[:ldash] + ' ' + pkgline[ldash + 1:]
        logging.debug("Found product [%s]", pkg)
        plist.append(pkg)
    logging.info("Completed retrieval of product details")
    return plist

def discover_freebsd(args, host):
    plist = []
    cmdarr = ["/usr/sbin/pkg info"]
    logging.info("Retrieving product details")
    pkgout = utils.run_cmd_on_host(args, host, cmdarr)

    begin = False
    for l in pkgout.splitlines():
        lsplit = l.split()
        pkgline = lsplit[0]
        ldash = pkgline.rfind('-')
        pkg = pkgline[:ldash] + ' ' + pkgline[ldash + 1:]
        logging.debug("Found product [%s]", pkg)
        plist.append(pkg)
    logging.info("Completed retrieval of product details")
    return plist

def discover_alpine(args, host):
    plist = []
    cmdarr = ["/sbin/apk list"]
    logging.info("Retrieving product details")
    pkgout = utils.run_cmd_on_host(args, host, cmdarr)

    begin = False
    for l in pkgout.splitlines():
        if l.startswith('WARNING:'):
            continue
        pkg = l.split()[0]
        ps = pkg.split('-')
        ver = ps[-2] + '-' + ps[-1]
        pkg = pkg.replace('-'+ver, '')
        pkg = pkg + ' ' + ver
        logging.debug("Found product [%s]", pkg)
        plist.append(pkg)
    logging.info("Completed retrieval of product details")
    return plist

def discover_macos(args, host):
    plist = []
    cmdarr = ["ls -la /Applications/"]
    logging.info("Retrieving product details")
    pkgout = utils.run_cmd_on_host(args, host, cmdarr)
    for l in pkgout.splitlines():
        l = l.strip()
        if l.startswith("d") and l.endswith(".") == False:
            tokens = l.split()
            app_name_tokens = tokens[8:] # there are 8 other fields apart from directory name
            app_name = " ".join(app_name_tokens)
            cmd = "defaults read \"/Applications/%s/Contents/Info.plist\" CFBundleShortVersionString" % app_name
            cmdarr = [cmd]
            out = utils.run_cmd_on_host(args, host, cmdarr, False)
            app_name = app_name[:-4]
            if out is not None:
                out = out.strip()
                app_name = app_name + " " + out
            plist.append(app_name)

    # Look for packages from any package manager
    
    # Home brew
    cmdarr = ["which brew"]
    out = utils.run_cmd_on_host(args, host, cmdarr, False)
    if out is not None and len(out.strip()) > 0:
        # Home brew is present
        cmdarr = ["brew list --versions"]
        pkgout = utils.run_cmd_on_host(args, host, cmdarr)
        if pkgout is not None:
            for l in pkgout.splitlines():
                plist.append(l.strip())
        
    logging.info("Completed retrieval of product details")
    return plist

def discover_rh(args, host):
    plist = []
    cmdarr = ["/usr/bin/yum list installed"]
    logging.info("Retrieving product details")
    yumout = utils.run_cmd_on_host(args, host, cmdarr)
    if yumout == None:
        logging.error("Error getting yum list output")
        return plist

    begin = False
    for l in yumout.splitlines():
        if 'Installed Packages' in l:
            begin = True
            continue
        if not begin:
            continue
        lsplit = l.split()
        pkg = lsplit[0]
        if len(lsplit) > 1:
            ver = lsplit[1]
        else:
            ver = ''
        pkgsp = pkg.split(".")
        if len(pkgsp) > 1:
            pkg = pkgsp[0]
            arch = pkgsp[1]
        else:
            pkg = pkgsp[0]
            arch = "noarch"

        if ':' in ver:
            ver = ver.split(':')[1]
        ver = ver + "." + arch
        logging.debug("Found product [%s %s]", pkg, ver)
        plist.append(pkg+' '+ver)
    logging.info("Completed retrieval of product details")
    return plist

def discover_suse(args, host):
    plist = []
    cmdarr = ["/bin/rpm -qa"]
    logging.info("Retrieving product details")
    rpmout = utils.run_cmd_on_host(args, host, cmdarr)

    for l in rpmout.splitlines():
        l = l.strip()
        l_tokens = l.split('-')
        tokens_len = len(l_tokens)
        pkg = "-".join(l_tokens[:-2])
        plist.append(pkg+' '+l_tokens[-2]+'-'+l_tokens[-1])
    logging.info("Completed retrieval of product details")
    return plist

def discover_ubuntu(args, host):
    plist = []
    cmdarr = ["/usr/bin/apt list --installed"]
    logging.info("Retrieving product details")
    yumout = utils.run_cmd_on_host(args, host, cmdarr)

    begin = False
    for l in yumout.splitlines():
        if 'Listing...' in l:
            begin = True
            continue
        if not begin:
            continue
        if l.strip() == '':
            continue
        lsplit = l.split()
        pkg = lsplit[0].split('/')[0]
        ver = lsplit[1]
        logging.debug("Found product [%s %s]", pkg, ver)
        plist.append(pkg+' '+ver)
    logging.info("Completed retrieval of product details")
    return plist

def discover(args):
    handle = args.handle
    token = args.token
    instance = args.instance

    host_list_file = args.remote_hosts_csv
    if host_list_file == None:
        host_list_file = args.host_list

    if host_list_file is not None:
        with open(host_list_file, mode='r') as csv_file:
            csv_reader = csv.DictReader(csv_file, quoting=csv.QUOTE_NONE, escapechar='\\')
            password = None
            remote_hosts = []
            for row in csv_reader:
                if not args.secure and row['userpwd'].startswith('__SECURE__:'):
                    if args.password is None:
                        if password is None:
                            password = getpass.getpass(prompt="Enter password: ")
                            password = password.encode()
                    else:
                        password = args.password.encode()
                    salt = base64.b64encode(password)
                    kdf = PBKDF2HMAC(
                            algorithm=hashes.SHA256(),
                            length=32,
                            salt=salt,
                            iterations=100000,
                            backend=default_backend())
                    key = base64.urlsafe_b64encode(kdf.derive(password))
                    f = Fernet(key)
                    try:
                        epass = row['userpwd'].replace('__SECURE__:','')
                        row['userpwd'] = f.decrypt(epass.encode('utf-8'))
                    except:
                        logging.error("Failed to decrypt login details for "+row['hostname'])
                        return None 
                elif row['userpwd'] != '' and not args.secure:
                    logging.warning('Unsecure login information in file. Use --secure to encrypt.')
                if '-' in row['hostname'] or '/' in row['hostname']: # IP range or CIDR is specified, then expand it
                    if '-' in row['hostname']:
                        iprange = row['hostname']
                        iprange = iprange.replace(' ','')
                        tokens = iprange.split('-')
                        if len(tokens) != 2 or len(tokens[0])==0 or len(tokens[1])==0:
                            logging.error("Skipping invalid range [%s]", row['hostname'])
                            continue
                        logging.info("Enumerating IPs based on specified range [%s]", iprange)
                        try:
                            if sys.version_info[0] < 3:
                                startip = ipaddress.IPv4Address(unicode(tokens[0]))
                                endip = ipaddress.IPv4Address(unicode(tokens[1]))
                            else:
                                startip = ipaddress.IPv4Address(tokens[0])
                                endip = ipaddress.IPv4Address(tokens[1])
                            cidrs = []
                            cidrs = [ipaddr for ipaddr in ipaddress.summarize_address_range(startip,endip)]
                        except Exception as e:
                            logging.error("Encountered exception: %s",e)
                            logging.error("Error converting IP range [%s] to CIDRs. Skipping it...", iprange)
                            continue
                        logging.info("Converted IP range [%s] to CIDRs %s", iprange, cidrs)
                    if '/' in row['hostname']:
                        logging.info("Enumerating IPs based on specified CIDR [%s]", row['hostname'])
                        try:
                            if sys.version_info[0] < 3:
                                cidrs = [ipaddress.ip_network(unicode(row['hostname'],"ascii"))]
                            else:
                                cidrs = [ipaddress.ip_network(row['hostname'])]
                        except Exception as e:
                            logging.error("Encountered exception: %s",e)
                            logging.error("Invalid CIDR [%s] specified. Skipping it...", row['hostname'])
                            continue
                    for cidr in cidrs:
                        for a in cidr:
                            trow = row.copy()
                            trow['hostname'] = str(a)

                            # Remove hard-coded asset ID and name for CIDR, as it will overwrite same asset
                            # These will based on host IP address automatically
                            trow['assetname'] = None
                        
                            remote_hosts.append(trow)
                            remote_hosts[-1]['remote'] = True
                            logging.info("Enumerated IP: %s", a)
                else:
                    remote_hosts.append(row)
                    remote_hosts[-1]['remote'] = True
        if args.secure:
            # secure the host list
            logging.info("Securing host list file")
            if args.password is None:
                pp1 = getpass.getpass(prompt="Enter password: ")
                pp2 = getpass.getpass(prompt="Re-enter password: ")
                if pp1 != pp2:
                    logging.info("Passwords don't match. Try again.")
                    return None
            else:
                pp1 = args.password
            password = pp1.encode()
            salt = base64.b64encode(password)
            kdf = PBKDF2HMAC(
                    algorithm=hashes.SHA256(),
                    length=32,
                    salt=salt,
                    iterations=100000,
                    backend=default_backend())
            key = base64.urlsafe_b64encode(kdf.derive(password))
            f = Fernet(key)
            # verify the key if possible
            with open(host_list_file, mode='r') as csv_file:
                csv_reader = csv.DictReader(csv_file, quoting=csv.QUOTE_NONE, escapechar='\\')
                for row in csv_reader:
                    try:
                        if row['userpwd'] != '' and row['userpwd'].startswith('__SECURE__:'):
                            epass = row['userpwd'].replace('__SECURE__:','')
                            epass  = f.decrypt(epass)
                    except:
                        logging.error("Invalid password")
                        logging.error("Please use the same password as was used previously to secure the file")
                        return None
            # create new file with secured information. This will be renamed subsequently
            new_csv_file = uuid.uuid4().hex
            new_csv_file = tempfile.gettempdir() + os.path.sep + new_csv_file + ".csv"
            if os.path.isfile(new_csv_file):
                os.remove(new_csv_file)
            try:
                # secure the new rows in the file
                with open(new_csv_file, mode='w') as csvfile:
                    fieldnames = ['hostname','userlogin','userpwd','privatekey','assetname']
                    writer = csv.DictWriter(csvfile, fieldnames=fieldnames, quoting=csv.QUOTE_NONE, escapechar='\\')
                    writer.writeheader()
                    for h in remote_hosts:
                        if h['userpwd'] != '' and not h['userpwd'].startswith('__SECURE__:'):
                            h['userpwd'] = '__SECURE__:'+f.encrypt(h['userpwd'].encode('utf-8')).decode('utf-8')
                        del h['remote']
                        writer.writerow(h)
            except Exception as err:
                logging.error("Unable to save secured CSV file")
                logging.error("%s", err)
                os.remove(new_csv_file)
                utils.tw_exit(1)
            os.remove(host_list_file)
            shutil.copyfile(new_csv_file, host_list_file)
            os.remove(new_csv_file)
            logging.info("Host list file secured")
            return None
        else:
            return discover_hosts(args, remote_hosts)
    else:
        host = { }
        host['assetid'] = utils.get_ip() if args.assetid is None or args.assetid.strip() == "" else args.assetid
        host['assetname'] = host['assetid'] if args.assetname is None or args.assetname.strip() == "" else args.assetname
        host['hostname'] = utils.get_ip()
        host['remote'] = False
        hosts = [ host ]
        return discover_hosts(args, hosts)

def discover_hosts(args, hosts):
    assets = []
    for host in hosts:
        asset = discover_host(args, host)
        if asset is not None:
            if args.no_ssh_audit == False:
                ssh_config_issues = []
                try:
                    ssh_config_issues = run_ssh_audit(args, asset['id'], host['hostname'])
                except Exception as e:
                    logging.error("Error parsing ssh audit: %s" % str(e))
                    logging.error(traceback.format_exc())
                asset['config_issues'] = ssh_config_issues
                if len(ssh_config_issues) != 0:
                    asset['tags'].append('SSH Audit')

            if args.no_host_benchmark == False:
                # Run host benchmark
                host_bm_issues = host_benchmark.run_host_benchmark(host, asset['id'], args)
                if len(host_bm_issues) > 0:
                    asset['tags'].append('Host Benchmark')
                    if asset.get('config_issues') is None:
                        asset['config_issues'] = host_bm_issues
                    else:
                        asset['config_issues'].extend(host_bm_issues)

            assets.append(asset)
    return assets

def discover_host(args, host):

    logging.info("Checking if host [%s] is reachable", host['hostname'])
    if not check_host_up(host):
        logging.error("Host is not reachable [%s]", host['hostname'])
        return None

    logging.info("Started inventory discovery for asset [%s]", host['hostname'])

    os = utils.get_os_release(args, host)
    if os is None:
        logging.error("Failed to identify OS for asset [%s]", host['hostname'])
        return None

    atype = utils.get_asset_type(os)
    if atype is None:
        logging.error("Could not determine asset type for asset [%s]", host['hostname'])
        return None

    asset_id = utils.get_unique_asset_id(args, host, atype)
    if asset_id is None:
        logging.warning("Could not get Identifier for Asset [%s]", host['hostname'])
        # Fallback to IP
        asset_id = host['hostname']
        logging.warning("Using Hostname/IP [%s] for Asset [%s] as Identifier", asset_id, host['hostname'])

    asset_name = host['assetname'] if host.get('assetname') is not None and len(host['assetname']) > 0 else host['hostname']

    asset_id = asset_id.replace(':','')
    asset_name = asset_name.replace('/','-')
    asset_name = asset_name.replace(':','-')

    plist = None
    if atype == 'CentOS' or atype == 'Red Hat' or atype == 'Amazon Linux' or atype == 'Oracle Linux' or atype == 'Rocky Linux':
        plist = discover_rh(args, host)
    elif atype == 'Ubuntu' or atype == 'Debian':
        plist = discover_ubuntu(args, host)
    elif atype == 'FreeBSD':
        plist = discover_freebsd(args, host)
    elif atype == 'OpenBSD':
        plist = discover_openbsd(args, host)
    elif atype == "Mac OS":
        plist = discover_macos(args, host)
    elif atype == "Alpine Linux":
        plist = discover_alpine(args, host)
    elif atype == "Suse":
        plist = discover_suse(args, host)

    if plist == None or len(plist) == 0:
        logging.error("Could not inventory asset [%s]", asset_id)
        return None

    asset_data = {}
    asset_data['id'] = asset_id
    asset_data['name'] = asset_name
    asset_data['type'] = atype
    asset_data['owner'] = args.handle
    asset_data['products'] = plist
    asset_tags = []
    asset_tags.append('OS_RELEASE:' + os)
    asset_tags.append(os)
    if atype != "Mac OS":
        asset_tags.append('Linux')
    asset_tags.append(atype)
    asset_data['tags'] = asset_tags
    utils.update_tool_run_record()

    plugin_processor.process_plugins(asset_data, host, args, '/')

    logging.info("Completed inventory discovery for asset [%s]", asset_id)

    return asset_data

def run_ssh_audit(args, assetid, ip):
    logging.info("Running ssh audit for "+ip)
    issue_list = []
    python_cmd = "python3"
    if sys.version_info.major < 3:
        python_cmd = "python"
    SSH_AUDIT_PATH = python_cmd + " " + os.path.dirname(os.path.realpath(__file__)) + '/ssh-audit.py'
    audit_out = ''
    try:
        cmd = SSH_AUDIT_PATH + ' -nv ' +ip
        cmdarr = [cmd]
        dev_null_device = open(os.devnull, "w")
        audit_out = subprocess.check_output(cmdarr, stderr=dev_null_device, shell=True)
        audit_out = audit_out.decode(args.encoding)
        dev_null_device.close()
    except subprocess.CalledProcessError as e:
        logging.error("Error running ssh audit: %s" % str(e))
        #logging.error(traceback.format_exc())
        return issue_list
    key_issues = {}
    recs = {}
    for l in audit_out.splitlines():
        if l.strip() == '':
            continue
        logging.debug("output line: %s" % l)
        larr = l.split()
        rtype = larr[0].strip()
        if rtype not in ['(cve)','(kex)','(key)','(enc)','(mac)','(rec)']:
            continue
        if ' [info] ' in l:
            continue
        if rtype == '(cve)':
            issue = { }
            issue['twc_id'] = 'ssh-audit: '+larr[1]
            title = l.split('CVSSv2:')[1].split(')')[1]
            cvss_score = l.split('CVSSv2:')[1].split(')')[0]
            issue['twc_title'] = 'ssh-audit: '+ larr[1]
            issue['details'] = larr[1]+'\n'+title+'\nCVSS Score '+cvss_score 
            issue['rating'] = utils.get_rating(cvss_score)
            issue['asset_id'] = assetid
            issue['object_id'] = ip
            issue['object_meta'] = ''
            issue['type'] = 'SSH'
            issue_list.append(issue)
        elif rtype in ['(kex)','(key)','(enc)','(mac)']:
            algo = larr[1]
            if algo not in key_issues:
                key_issues[algo] = {}
                key_issues[algo]['type'] = larr[0].replace('(','').replace(')','')
                if '--' in l:
                    rating = l.split('--')[1].split()[0]
                    detail = l.split('--')[1].split(']')[1].strip()
                elif '`-' in l:
                    rating = l.split('`-')[1].split()[0]
                    detail = l.split('`-')[1].split(']')[1].strip()
                else:
                    continue
                if rating == '[fail]':
                    rating = '4'
                else:
                    rating = '3'
                key_issues[algo]['rating'] = rating 
                title = ""
                if rtype == '(kex)':
                    title = 'ssh-audit: Unsafe key exchange - '+algo
                elif rtype == '(key)':
                    title = 'ssh-audit: Unsafe key - '+algo
                elif rtype == '(mac)':
                    title = 'ssh-audit: Unsafe mac algorithm - '+algo
                elif rtype == '(enc)':
                    title = 'ssh-audit: Unsafe encryption - '+algo
                key_issues[algo]['title'] = title 
                key_issues[algo]['details'] = detail 
            else:
                if '--' in l:
                    detail = l.split('--')[1].split(']')[1].strip()
                else:
                    detail = l.split('`-')[1].split(']')[1].strip()
                if 'details' in key_issues[algo]:
                    key_issues[algo]['details'] = key_issues[algo]['details'] + '\n'+ detail
                else:
                    key_issues[algo]['details'] = detail
        elif rtype == '(rec)':
            algo = larr[1][1:]
            if algo not in key_issues:
                key_issues[algo] = {}
                key_issues[algo]['details'] = ''
            reco = 'Recommendation: '+l.split('--')[1].strip()
            key_issues[algo]['details'] = key_issues[algo]['details'] + '\n'+ reco 
    for k in key_issues:
        if 'type' not in key_issues[k] or 'title' not in key_issues[k]:
            continue
        issue = {}
        issue['twc_id'] = 'ssh-audit-'+key_issues[k]['type'] + '-' + k 
        issue['twc_title'] = key_issues[k]['title']
        issue['details'] = key_issues[k]['details']
        issue['rating'] = key_issues[k]['rating']
        issue['object_id'] = k
        issue['asset_id'] = assetid
        issue['object_meta'] = '' 
        issue['type'] = 'SSH'
        issue_list.append(issue)

    logging.info("ssh audit completed "+ip)
    return issue_list

def get_inventory(args):
    return discover(args)



import sys
sys.path.append('../twigs')
import os
import subprocess
import tempfile
import shutil
import hashlib
import tarfile
import logging
import json

from twigs import utils

# Note this error routine assumes that the file was read-only and hence could not be deleted
def on_rm_error( func, path, exc_info):
    os.chmod( path, stat.S_IWRITE )
    os.unlink( path )

def process_line(line, rating, asset_id):
    bundled_recommendations_path = os.path.dirname(os.path.realpath(__file__)) + os.sep + 'recommendations.json'
    rf = open(bundled_recommendations_path, 'r')
    rdata = rf.read()
    rf.close()
    rdict = json.loads(rdata)
    line_fields = line.split('|')
    category = line_fields[0]
    category_url = "https://cisofy.com/lynis/controls/" + category
    description = line_fields[1]
    suggestion = line_fields[2] if line_fields[2] != '-' else ''
    extra = line_fields[3] if line_fields[3] != '-' else ''
    config_issue = { }
    twc_id_str = category + '-' + description
    if len(suggestion) > 0:
        twc_id_str = twc_id_str + '-' + suggestion
    twc_id_hash = hashlib.md5(twc_id_str.encode())
    config_issue['twc_id'] = twc_id_hash.hexdigest()
    config_issue['asset_id'] = asset_id
    config_issue['twc_title'] = "[%s] %s" % (category, description)
    if rdict.get(category):
        details = rdict[category] 
    else:
        details = suggestion + '\n\nFor additional information refer:\n' + category_url + '\n'
        details = details.strip()
    config_issue['details'] = details
    config_issue['type'] = 'Host Benchmark'
    config_issue['rating'] = rating
    config_issue['object_id'] = ''
    config_issue['object_meta'] = ''
    return config_issue

def process_report(report_file, asset_id):
    bundled_recommendations_path = os.path.dirname(os.path.realpath(__file__)) + os.sep + 'recommendations.json'
    config_issues = []
    with open(report_file, 'r') as rpt_file:
        while True:
            line = rpt_file.readline()
            if not line:
                break
            line = line.strip()
            if line.startswith('warning[]='):
                config_issue = process_line(line[10:], '4', asset_id)
                config_issues.append(config_issue)
            elif line.startswith('suggestion[]='):
                config_issue = process_line(line[13:], '2', asset_id)
                config_issues.append(config_issue)
    return config_issues

def run_host_benchmark(host, asset_id, args):
    logging.info("Running host benchmark for [%s]. This may take some time...", host['hostname'])
    args_encoding = args.encoding if args is not None else 'latin-1'
    config_issues = []
    bundled_tar_path = os.path.dirname(os.path.realpath(__file__)) + os.sep + 'lynis-3.1.1.tar.gz'
    local_temp_dir = tempfile.mkdtemp()

    # Untar tar to tmp directory
    with tarfile.open(bundled_tar_path, 'r', format=tarfile.PAX_FORMAT) as tf:
        tf.extractall(path=local_temp_dir)

    # Create local tmp directory
    if host['remote'] == False:
        extract_tar_path = local_temp_dir
    else:
        # SSH and create tmp directory on remote box
        cmdarr = ['mktemp -d']
        remote_temp_dir = utils.run_cmd_on_host(args, host, cmdarr)

        remote_temp_dir = remote_temp_dir.strip()
        # SCP tar contents to remote box
        utils.scp_put_file(host, local_temp_dir + os.sep + 'lynis-3.1.1', remote_temp_dir)
        extract_tar_path = remote_temp_dir

    # Run host benchmark tool
    cmdarr = ['cd ' + extract_tar_path + os.sep + 'lynis-3.1.1 && chmod 640 include/* && ./lynis audit system --quiet --logfile ../tw_lynis.log --report-file ../tw_lynis_report.dat && cd -']
    utils.run_cmd_on_host(args, host, cmdarr)

    if host['remote'] == True:
        # SCP report back to local tmp directory
        utils.scp_get_file(host, remote_temp_dir + os.sep + 'tw_lynis_report.dat', local_temp_dir)
        # Remove remote tmp directory
        cmdarr = ['rm -rf '+ remote_temp_dir]
        utils.run_cmd_on_host(args, host, cmdarr)

    # Process report from local tmp directory
    config_issues = process_report(local_temp_dir + os.sep + 'tw_lynis_report.dat', asset_id)

    # Remove local tmp directory
    shutil.rmtree(local_temp_dir, onerror = on_rm_error)

    logging.info("Completed host benchmark for [%s]", host['hostname'])

    return config_issues

"""
# Test local host benchmarks
local_host = { 'remote': False, 'hostname':'127.0.0.1'}
ci = run_host_benchmark(local_host, "test_asset", None)
print(ci)
"""

"""
# Test remote host benchmarks
remote_host = {
    'remote': True,
    'hostname':'hostname.somecompany.io',
    'userlogin':'userlogin',
    'userpwd':'userpwd',
    'privatekey': '',
    'assetname': 'dummy'
}
ci = run_host_benchmark(remote_host, "test_asset", None)
print(ci)
"""



import sys
import json
import os
import subprocess
import requests
import re
import logging
import datetime

from . import utils

g_encoding = None
g_required_extensions = ["account", "log-analytics"]
gTenantId = None

def set_encoding(encoding):
    global g_encoding
    g_encoding = encoding

def get_encoding():
    global g_encoding
    return g_encoding

def run_az_cmd(cmd, suppress_errors=False):
    cmd = 'az ' + cmd + ' --output json --only-show-errors'
    if suppress_errors:
        cmd = cmd + ' 2>/dev/null'
    try:
        logging.debug("Running cmd [%s]", cmd)
        cmd_output = subprocess.check_output([cmd], shell=True, stdin=None, stderr=None)
        cmd_output = cmd_output.decode(get_encoding())
        ret_json = json.loads(cmd_output)
    except subprocess.CalledProcessError:
        if suppress_errors:
            logging.debug("Error running az command [%s]", cmd)
            return None
        else:
            logging.error("Error running az command [%s]", cmd)
            utils.tw_exit(1)
    except ValueError:
        if suppress_errors:
            logging.debug("Error parsing JSON output for az command [%s]: %s", cmd, cmd_output)
            return None
        else:
            logging.error("Error parsing JSON output for az command [%s]: %s", cmd, cmd_output)
            utils.tw_exit(1)
    return ret_json

def check_required_az_extensions():
    global g_required_extensions
    installed_extensions = []
    missing_extensions = []
    installed_extensions_json = run_az_cmd('extension list')
    for ie in installed_extensions_json:
        installed_extensions.append(ie['name'])
    for req_ext in g_required_extensions:
        if req_ext not in installed_extensions:
            missing_extensions.append(req_ext)
    if len(missing_extensions) > 0:
        logging.error("Please install following Azure CLI extensions for discovery: %s", str(missing_extensions))
        logging.error("Use [az extension add --name <extension_name>] command to install missing extensions")
        utils.tw_exit(1)

def get_tenant_id():
    global gTenantId
    if gTenantId is not None:
        return gTenantId
    rjson = run_az_cmd('account tenant list')
    gTenantId = rjson[0]['tenantId']
    return gTenantId

# Main entry point
def get_inventory(args):
    set_encoding(args.encoding)
    check_required_az_extensions()
    assets = retrieve_inventory(args)
    return assets

def convert_to_datetime(datetime_str):
    if '.' in datetime_str:
        # Known behavior with Azure TimeGenerate values wherein at times it includes microseconds and that too with 7 digits (instead of 6 digits)
        datetime_str = datetime_str.split('.')[0]
        return datetime.datetime.strptime(datetime_str,  "%Y-%m-%dT%H:%M:%S")
    else:
        return datetime.datetime.strptime(datetime_str,  "%Y-%m-%dT%H:%M:%SZ")

def is_product_removed(rpt, assetid, resource_id, pnv, last_reported_on):
    tasset = rpt.get(resource_id)
    if tasset is None:
        return False
    removed_on  = rpt[resource_id].get(pnv)
    if removed_on is None:
        return False
    else:
        removed_on = convert_to_datetime(removed_on)
        last_reported_on = convert_to_datetime(last_reported_on)
        if removed_on > last_reported_on:
            logging.debug("Product [%s] removed on asset [%s] with resource_id [%s]", pnv, assetid, resource_id)
            logging.debug("ConfigurationData product last_reported_on %s", last_reported_on)
            logging.debug("ConfigurationChange product removed_on %s", removed_on)
            return True
    return False

def prepare_removed_product_tracker(workspace_id):
    logging.info("Processing configuration change data retrieved from Azure...")
    # Removed Product Tracker - this is a 2 level dict as assetid -> product -> removedtime 
    rpt = { }
    rjson = run_az_cmd("monitor log-analytics query -w '%s' --analytics-query 'ConfigurationChange | where ConfigChangeType == \"Software\" and ChangeCategory == \"Removed\" | summarize arg_max(TimeGenerated, *) by _ResourceId, Publisher, SoftwareName, SoftwareType | project _ResourceId, SoftwareName, Previous, SoftwareType, TimeGenerated'" % workspace_id)
    logging.debug("ConfigurationChange data: %s", rjson)
    for item in rjson:
        if item['SoftwareType'] not in ['Application', 'Package']:
            # currently we only handle software uninstallations
            continue
        pname = item['SoftwareName']
        pversion = item['Previous']
        # Azure Monitoring Agent bug - version has "(none):1.4.6-1.el8" for Linux packages
        pversion = pversion[7:] if item['SoftwareType'] == 'Package' and pversion.startswith('(none):') else pversion
        pnv = pname + ' ' + pversion
        resource_id = item['_ResourceId']
        if resource_id not in rpt:
            rpt[resource_id] = { }
        rpt[resource_id][pnv] = item['TimeGenerated']
    logging.debug("Removed product tracker dict: %s", rpt)
    return rpt

def parse_inventory(args, data, rpt):
    logging.info("Processing inventory retrieved from Azure...")
    hosts = []
    assets = []
    asset_map = {}
    not_running_vms = {}
    all_assets = { }
    for item in data:
        all_assets[item['_ResourceId']] = item['Computer']
        logging.debug("Parsing inventory from data below:\n%s", json.dumps(item, indent=2))
        host = item['Computer']
        resource_id = item['_ResourceId']
        publisher = item['Publisher']

        # If VM is known to be not running, then skip it
        if not_running_vms.get(resource_id) == 1:
            continue

        if host not in hosts  and publisher != '0':
            logging.debug("Found new asset - host [%s] resource_id [%s]", host, resource_id)
            patches = []
            products = []
            asset_map = {}
            asset_map['owner'] = args.handle
            asset_map['host'] = host
            asset_map['id'] = item['VMUUID']
            asset_map['name'] = host
            asset_map['tags'] = [ ]
            asset_map['patch_tracker'] = { } # To help remove duplicate patches
            asset_map['resource_id'] = resource_id
            if item['SoftwareType'] in ['Update', 'Patch']: #ApplicationType for MS patches
                patch = parse_patch(item)
                if patch is not None:
                    patches.append(patch)
                    asset_map['patch_tracker'][patch['id']] = patch['id']
            if item['SoftwareType'] == 'Package' or item['SoftwareType'] == 'Application': #ApplicationType for Linux packages
                pname = item['SoftwareName']
                pversion =  item['CurrentVersion']
                # Azure Monitoring Agent bug - version has "(none):1.4.6-1.el8" for Linux packages
                pversion = pversion[7:] if item['SoftwareType'] == 'Package' and pversion.startswith('(none):') else pversion
                pnv = pname + ' ' + pversion
                if not is_product_removed(rpt, asset_map['id'], asset_map['resource_id'], pnv, item['TimeGenerated']):
                    products.append(pnv)
            asset_map['products'] = products
            asset_map['patches'] = patches
            vm_running, os, os_version, sub_id, tags = get_vm_details(host, resource_id)
            if vm_running == False:
                # skip vm's which are not running
                not_running_vms[resource_id] = 1
                continue
            asset_map['type'] = get_os_type(os)
            if len(asset_map['type']) > 0:
                asset_map['tags'].append(asset_map['type'])
            if len(tags) > 0:
                asset_map['tags'].extend(tags)
            if asset_map['type'] == 'Windows':
                asset_map['tags'].append('OS_RELEASE:' + os)
                asset_map['tags'].append(os)
                asset_map['tags'].append('OS_VERSION:' + os_version)
            else:
                asset_map['tags'].append('OS_RELEASE:%s %s' % (os, os_version))
                asset_map['tags'].append('%s %s' % (os, os_version))
            if args.enable_tracking_tags == True:
                asset_map['tags'].append("SOURCE:Azure:Tenant:" + get_tenant_id())
                asset_map['tags'].append("Azure:Tenant:" + get_tenant_id())
                asset_map['tags'].append("SOURCE:Azure:Subscription:" + sub_id)
                asset_map['tags'].append("Azure:Subscription:" + sub_id)
            else:
                asset_map['tags'].append("SOURCE:Azure")
                asset_map['tags'].append("Azure")
            assets.append(asset_map)
            hosts.append(host)
        else:
            for asset in assets:
                if asset['resource_id'] == resource_id:
                    if item['SoftwareType'] in ['Update', 'Patch']: #ApplicationType for MS patches
                        patch = parse_patch(item)
                        if patch is not None and asset['patch_tracker'].get(patch['id']) is None:
                            asset['patches'].append(patch)
                            asset['patch_tracker'][patch['id']] = patch['id']
                    if item['SoftwareType'] == 'Package' or item['SoftwareType'] == 'Application': #ApplicationType for Linux packages
                        pname = item['SoftwareName']
                        pversion =  item['CurrentVersion']
                        # Azure Monitoring Agent bug - version has "(none):1.4.6-1.el8" for Linux packages
                        pversion = pversion[7:] if item['SoftwareType'] == 'Package' and pversion.startswith('(none):') else pversion
                        pnv = pname + ' ' + pversion
                        if not is_product_removed(rpt, asset['id'], asset['resource_id'], pnv, item['TimeGenerated']):
                            asset['products'].append(pnv)

    # Remove the additional fields 'patch_tracker' (added to avoid duplicate patches) & 'resource_id'
    for asset in assets:
        asset.pop('patch_tracker', None)
        asset.pop('resource_id', None)
    logging.debug("Total assets reported: %s", len(all_assets))
    logging.debug("Assets with s/w packages and patches: %s", len(assets))
    logging.debug("Not running VMs: %s", len(not_running_vms))
    return assets

def parse_patch(item):
    patch_id = re.findall(r'(KB[0-9]+)', item['SoftwareName'])
    if len(patch_id) == 0:
        return None
    patch = {}
    patch['url'] = ''
    patch['id'] = patch_id[0]
    patch['product'] = ''
    patch['description'] = item['SoftwareName']
    return patch
            
def get_os_type(ostype):
    asset_type = ''
    if ostype is None:
        asset_type = ''
    elif 'Microsoft' in  ostype or 'Windows' in ostype:
        asset_type = 'Windows'
    elif 'Red Hat' in ostype or 'redhat' in ostype or 'rhel' in ostype or 'RHEL' in ostype:
        asset_type = 'Red Hat'
    elif 'Ubuntu' in ostype or 'ubuntu' in ostype:
        asset_type = 'Ubuntu'
    elif 'CentOS' in ostype or 'centos' in ostype:
        asset_type = 'CentOS'
    elif 'Oracle' in ostype or 'oracle' in ostype:
        asset_type = "Oracle Linux"
    logging.debug("Mapped OS [%s] to Asset Type [%s]", ostype, asset_type)
    return asset_type

def retrieve_inventory(args):
    email = args.handle
    workspace_id = args.azure_workspace
    rpt = prepare_removed_product_tracker(workspace_id)
    rjson = run_az_cmd("monitor log-analytics query -w '%s' --analytics-query 'ConfigurationData | where ConfigDataType == \"Software\" | summarize arg_max(TimeGenerated, *) by _ResourceId, Publisher, SoftwareName, SoftwareType | project Computer, _ResourceId, VMUUID, ConfigDataType, Publisher, SoftwareName, SoftwareType, CurrentVersion, TimeGenerated'" % workspace_id)
    return parse_inventory(args, rjson, rpt)

def is_vm_running(vm_json):
    statuses = vm_json.get('statuses')
    if statuses is None:
        return False
    for status in statuses:
        status_code = status.get('code')
        if status_code is not None and status_code.startswith('PowerState/'):
            if status_code.split('/')[1] == "running":
                return True
            else:
                return False
    return False

# Try to get OS details, subscription Id and tags for given VM
def get_vm_details(host, resource_id):
    logging.debug("Getting OS details for host [%s] resource_id [%s]", host, resource_id)
    tags = []
    rjson = run_az_cmd("vm get-instance-view --ids '%s'" % resource_id, True)
    if rjson is None:
        # VM not found (probably deleted)
        return False, None, None, None, tags
    ijson = rjson['instanceView']
    # At times VM is marked as running but osName details are not yet populated in instanceView JSONand in such cases it is best to skip the VM in this discovery for now.
    if is_vm_running(ijson) and ijson.get('osName') is not None:
        rid = rjson['id']
        rid_tokens = rid.split('/')
        sub_id = rid_tokens[2]
        if rjson.get('tags') is not None:
            vm_tags = rjson['tags']
            for key in list(vm_tags.keys()):
                tag_name = key
                tag_value = vm_tags[key]
                tags.append(tag_name + ':' + tag_value)
        return True, ijson.get('osName'), ijson.get('osVersion'), sub_id, tags
    else:
        return False, None, None, None, tags




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import sys

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['boto3', 'setuptools', 'requests', 'requirements_parser', 'pysnow', 'ipaddress==1.0.22', 'pefile==2019.4.18', 'paramiko==2.6.0', 'cryptography==3.3.2', 'toml==0.10.2', 'pyvmomi==7.0.3', 'scp==0.13.3', 'pyyaml>5.1', 'psutil', 'ipaddress', 'oci-cli==3.43.1', 'pytz', 'pywinrm==0.5.0']
if sys.platform != 'win32':
    requirements.append('python-crontab==2.5.1')

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="Paresh Borkar",
    author_email='opensource@threatwatch.io',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="ThreatWorx Information Gathering Script",
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='twigs',
    name='twigs',
    packages=find_packages(include=['twigs', 'twigs.dast_plugins', 'twigs.azure_cis_tool', 'twigs.gcp_cis_tool', 'twigs.oci_cis_tool']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/threatworx/twigs',
    version='1.2.73',
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'twigs=twigs.twigs:main',
        ],
    },
)



import sys
import json
import re
import logging
import requests
import pysnow

from . import utils as utils

def get_product_version(asset_display_name):
    temp_product_details = re.findall(r'(.*)\((.*)\)',asset_display_name)
    return temp_product_details[0][1]

def find_host(host_cache, all_hosts, host_id):
    host_slot = host_cache.get(host_id)
    if (host_slot == None):
        return None
    return all_hosts[host_slot]

def build_host_cache(all_hosts):
    host_cache = {}
    host_count = 0
    for host in all_hosts:
        host_cache[host['sys_id']] = host_count
        host_count = host_count + 1 
    return host_cache

def get_asset_inventory(args):
    access_token = None # only set for OAuth flow
    if args.snow_client_id is not None:
        if args.snow_client_secret is None or len(args.snow_client_secret.strip()) == 0:
            logging.error("Client secret is required for OAuth")
            utils.tw_exit(1)
        logging.info("Getting access token using OAuth")
        snow_oauth_url = "https://%s.service-now.com/oauth_token.do" % args.snow_instance
        oauth_payload = "grant_type=client_credentials&client_id=%s&client_secret=%s" % (args.snow_client_id, args.snow_client_secret)
        headers = {
            'Content-Type': "application/x-www-form-urlencoded",
            'Cache-Control': "no-cache"
        }
        response = requests.request("POST", snow_oauth_url, data=oauth_payload, headers=headers)
        if response.status_code == 200:
            access_token = response.json()['access_token']
            logging.info("Obtained access token")
        else:
            logging.error("OAuth failed, unable to get access token")
            logging.debug(response.text)
            utils.tw_exit(1)

    if args.snow_user is not None:
        if args.snow_user_pwd is None or len(args.snow_user_pwd.strip()) == 0:
            logging.error("User password is required for user authentication")
            utils.tw_exit(1)
        client = pysnow.Client(instance=args.snow_instance, user=args.snow_user, password=args.snow_user_pwd)

    logging.info("Getting asset inventory from ServiceNow (this may take some time) ...")
    if access_token is None:
        snow_base_api_path = '/table/'
        snow_api_path = snow_base_api_path + 'alm_entitlement_asset'
        alm_entitlement_assets = client.resource(api_path=snow_api_path)
        alm_entitlement_assets_response = alm_entitlement_assets.get(query={'sys_class_name':'alm_entitlement_asset'})
        alm_entitlement_assets_list = alm_entitlement_assets_response.all()
        # get all hosts and add these to the cache
        snow_hosts_api_path = snow_base_api_path + 'cmdb_ci_computer'
        all_hosts_query = client.resource(api_path=snow_hosts_api_path)
        all_hosts_response = all_hosts_query.get(query={})
        all_hosts = all_hosts_response.all()
    else:
        api_url = "http://%s.service-now.com/api/now/table/alm_entitlement_asset" % args.snow_instance
        headers = {"Authorization": "Bearer "+access_token, "Content-Type":"application/json","Accept":"application/json"}
        response = requests.post(api_url, headers=headers, data="{'sys_class_name':'alm_entitlement_asset'}" )
        if response.status_code != 200:
            logging.error("Unable to get software inventory")
            logging.debug("REST API call to get alm_entitlement_asset rows failed, response details below")
            logging.debug(response.text)
            utils.tw_exit(1)
        alm_entitlement_assets_list = response.json()['result']
        api_url = "http://%s.service-now.com/api/now/table/cmdb_ci_computer" % args.snow_instance
        headers = {"Authorization": "Bearer "+access_token, "Content-Type":"application/json","Accept":"application/json"}
        response = requests.post(api_url, headers=headers, data='{}' )
        if response.status_code != 200:
            logging.error("Unable to get asset information")
            logging.debug("REST API call to get cmdb_ci_computer rows failed, response details below")
            logging.debug(response.text)
            utils.tw_exit(1)
        all_hosts = response.json()['result']

    host_cache = build_host_cache(all_hosts)
    logging.info("Processing asset information....")
    hosts = {}
    for alm_entitlement_asset in alm_entitlement_assets_list:
        product_version = get_product_version(alm_entitlement_asset['display_name'])
        host_id = alm_entitlement_asset['allocated_to']['value']
        host = hosts.get(host_id)
        if (host == None):
            # Get the host entry from SNow
            snow_host = find_host(host_cache, all_hosts, host_id)
            if (snow_host == None):
                # error condition
                logging.error("Error: Could not find host:", host_id)
                continue
 
            # create new host entry and add it to the dictionary
            host = {}
            host['id'] = host_id
            # set asset name
            if len(snow_host['ip_address']) != 0:
                host['name'] = snow_host['ip_address']
            elif len(snow_host['fqdn']) != 0:
                host['name'] = snow_host['fqdn']
            elif len(snow_host['name']) != 0:
                host['name'] = snow_host['name']
            else:
                host['name'] = host['id']
            host['type'] = utils.get_asset_type(snow_host['os'])
            if host['type'] is None:
                host['type'] = 'Other'
            host['owner'] = args.handle
            products = []
            products.append(snow_host['os'])
            products.append(product_version)
            host['products'] = products
            asset_tags = []
            if args.enable_tracking_tags == True:
                asset_tags.append('SOURCE:ServiceNow:' + args.snow_instance)
            else:
                asset_tags.append('SOURCE:ServiceNow')
            asset_tags.append('ServiceNow')
            asset_tags.append(snow_host['asset_tag'])
            asset_tags.append('OS_RELEASE:' + snow_host['os'])
            asset_tags.append(snow_host['os'])
            if host['type'] != 'Other':
                asset_tags.append(host['type'])
            host['tags'] = asset_tags
            hosts[host_id] = host
        else:
            host['products'].append(product_version)

    # Convert dict of hosts to list of assets
    assets = [ h for h in hosts.values() ]
    return assets
 
def get_inventory(args):
    assets = get_asset_inventory(args)
    logging.info("Total %s assets found in inventory...",str(len(assets)))
    return assets




import sys
import platform
import os
import subprocess
import logging
import psutil
import ipaddress
import socket
import json
import re
import requests
from xml.dom.minidom import parse, parseString
from distutils.version import LooseVersion 
import csv
from . import linux
import shutil

NMAP_default = "/usr/bin/nmap"
NMAP = shutil.which("nmap")

SNMPWALK = shutil.which('snmpwalk')

if NMAP is None:
    NMAP = NMAP_default
NSE_PATH = os.path.dirname(os.path.realpath(__file__)) + '/nse/'

NMAP_HTTP_PORTS = ['80','443','6443','8080','8443','2181'] 
NSE_APACHE_PATH  = "+/"+os.path.dirname(os.path.realpath(__file__)) + '/nse/apache/'
NSE_HTTP_PATH = "+/"+os.path.dirname(os.path.realpath(__file__)) + '/nse/http/'
NSE_HTTP_SCRIPTS = ['+http-generator','+http-wordpress-enum','+http-apache-server-status','+http-server-header','+http-php-version',NSE_HTTP_PATH,NSE_APACHE_PATH]

NMAP_DB_PORTS = ['9200','9300','27017','27018','27019','3306','7000','7001','9042','7199','523','445','1443','6379','1521']
NSE_DB_PATH  = "/"+os.path.dirname(os.path.realpath(__file__)) + '/nse/database/'
NSE_DB_SCRIPTS = [NSE_DB_PATH,'mongodb-info','mysql-info','cassandra-info','db2-das-info','ms-sql-info','redis-info','oracle-tns-version']

NMAP_PRINTERS_PORTS = ['80','161','443','9100','U:161']
NSE_PRINTERS_PATH  = "/"+os.path.dirname(os.path.realpath(__file__)) + '/nse/printers/'
NSE_PRINTERS_SCRIPTS = [NSE_PRINTERS_PATH]

NMAP_CCTV_PORTS = ['21','80','161','443','8080','8443','4321', '37777', '9000', '10554', '5985','9100','5060']
NSE_CCTV_PATH  = "/"+os.path.dirname(os.path.realpath(__file__)) + '/nse/cctv/'
NSE_CCTV_SCRIPTS = [NSE_CCTV_PATH]

NSE_OTHER_PATH =  "/"+os.path.dirname(os.path.realpath(__file__)) + '/nse/other/' 

def nmap_exists():
    return NMAP and os.access(NMAP, os.X_OK)

def build_snmp_walk_cmd(args, addr):
    cmd = SNMPWALK
    if args.snmp_security_name:
        cmd = cmd + ' -v3 -u '+args.snmp_security_name
    else:
        cmd = cmd + ' -v1 '
    cmd = cmd + ' -c '+args.snmp_community + ' ' + addr
    return cmd

def get_snmp_oid_value(args, snmpwalk, oid):
    out = None
    try:
        logging.debug("snmpwalk command: " + snmpwalk + ' ' + oid)
        out = subprocess.check_output([snmpwalk+' '+oid], shell=True)
        out = out.decode(args.encoding)
    except subprocess.CalledProcessError:
        logging.error("Error running snmpwalk command")
        return out
    try:
        if ':' in out:
            out = out.strip().split(':')[1].replace('"','')
        else:
            out = out.strip().split('=')[1].replace('"','')
    except Exception as e:
        logging.error("Exception processing snmp oid walk output")
        return None
    logging.debug("snmpwalk output: " + out)
    return out

def discover(args):
    handle = args.handle
    token = args.token
    instance = args.instance

    hosts = args.hosts

def get_private_ip_cidrs():
    """
    Returns a list of class A, B, or C private IP CIDRs visible on the local host.
    """
    private_cidrs = []
    for interface in psutil.net_if_addrs().values():
        for address in interface:
            if address.family == socket.AF_INET:
                ip_address = ipaddress.IPv4Address(address.address)
                if ip_address.is_private:
                    if ipaddress.IPv4Network(ip_address).is_private:
                        if ip_address.is_private and \
                                (ipaddress.IPv4Address('10.0.0.0') <= ip_address <= ipaddress.IPv4Address('10.255.255.255') or
                                ipaddress.IPv4Address('172.16.0.0') <= ip_address <= ipaddress.IPv4Address('172.31.255.255') or
                                ipaddress.IPv4Address('192.168.0.0') <= ip_address <= ipaddress.IPv4Address('192.168.255.255')):
                            private_cidrs.append(str(ipaddress.IPv4Network(ip_address).with_prefixlen))
    return private_cidrs

# get the most accurate os_class object from host xml
def get_best_os_class(host):
    best_os_class = None
    host_os_classes = host.getElementsByTagName("osclass")
    if host_os_classes is not None:
        accuracy = -1
        for host_os_class in host_os_classes:
            host_acc = host_os_class.getAttribute("accuracy")
            if host_acc == "":
                continue
            host_acc = int(host_acc)
            if host_acc >= accuracy:
                best_os_class = host_os_class
                accuracy = host_acc
    return best_os_class

# get the os name from smb-os-discovery scrip output if present
def get_smb_os(host):
    smb_os_name = None
    scripts = host.getElementsByTagName("script")
    for s in scripts:
        if s.getAttribute('id') == 'smb-os-discovery':
            elems = s.getElementsByTagName('elem')
            for elem in elems:
               if elem.getAttribute('key') == 'os':
                  smb_os_name = " ".join(t.nodeValue for t in elem.childNodes if t.nodeType == t.TEXT_NODE)
                  break
    return smb_os_name

# get the most confident os type from services section
def get_os_from_services(host):
    conf = -1
    os_type = None
    services = host.getElementsByTagName("service")
    for service in services:
        ostype = service.getAttribute("ostype")
        if ostype != "":
            if int(service.getAttribute("conf")) > conf:
                os_type = ostype
                conf = int(service.getAttribute("conf"))
    return os_type

def get_os_type(host, products):
    # check the smb output to see if this is a windows host
    if get_smb_os(host):
        logging.debug("Found os_type [Windows] from SMB")
        return 'Windows'

    # check the os class
    os_class = get_best_os_class(host)
    if os_class:
        os_type = os_class.getAttribute("osfamily")
        logging.debug("Found os_type [%s] from os class", os_type)
        return os_type

    # check os type from services
    service_os = get_os_from_services(host)
    if service_os:
        logging.debug("Found os_type [%s] from services", service_os)
        return service_os

    # guess os type from products
    for product in products:
        if 'microsoft windows' in product:
            logging.debug("Found os_type [Windows] from products")
            return 'Windows', os_name
        if 'linux linux kernel' in product:
            logging.debug("Found os_type [Linux] from products")
            return 'Linux', os_name

    logging.debug("Unable to determine os_type...assuming [Other]")
    return 'Other'

def is_port_open(port):
    port_state = port.getElementsByTagName("state")[0]
    if port_state is not None and port_state.getAttribute('state') == "open":
        return True
    return False

def create_open_ports_issues(ports_in_use_dict, asset_id):
    open_port_issues = []
    for key in sorted(ports_in_use_dict.keys()):
        port_details = ports_in_use_dict[key]
        issue = {}
        issue['twc_id'] = 'OPEN_PORT_' + port_details['port']
        issue['twc_title'] = "Open Port [%s] detected" % port_details['port']
        summary = "Port [%s] is open for protocol [%s] on the host" % (port_details['port'], port_details['protocol'])
        if port_details['service'] is not None:
            summary = summary + " running [%s] service" % port_details['service']
        if port_details['product'] != "":
            summary = summary + " for product [%s]" % port_details['product']
        issue['details'] = summary
        issue['rating'] = '1'
        issue['object_id'] = port_details['port']
        issue['asset_id'] = asset_id
        issue['object_meta'] = ''
        issue['type'] = 'Open Ports'
        open_port_issues.append(issue)
    return open_port_issues

def create_nmap_cmd (args):
    ports = [] 
    scripts = [] 
    os = ""
    vflag = ""
    if 'default' in args.services: 
        vflag = " -sV "
    if "database" in args.services:
        ports += NMAP_DB_PORTS
        scripts += NSE_DB_SCRIPTS
    if "web" in args.services:
        ports += NMAP_HTTP_PORTS
        scripts += NSE_HTTP_SCRIPTS
    if "os" in args.services: 
        vflag = " -sV "
        os = " -O "
        scripts += ['smb-os-discovery']
        ports += ['1-100'] 
    if "vmware" in args.services:
        scripts += ['vmware-version']
        ports += ['443'] 
    if "printers" in args.services:
        ports += NMAP_PRINTERS_PORTS
        scripts += NSE_PRINTERS_SCRIPTS
    if "cctv" in args.services:
        ports += NMAP_CCTV_PORTS
        scripts += NSE_CCTV_SCRIPTS
    if "snmp" in args.services:
        ports = ['161']
        vflag = " -sU "
 
    cmd = NMAP + vflag + ' -Pn -oX - -T ' + args.timing + os
    if len(ports) != 0:
        cmd += ' -p' + ','.join(list(set(ports)))
    if args.extra_ports:
        cmd += ','+args.extra_ports
    if len(scripts) != 0:
        cmd += ' --script '+','.join(list(set(scripts)))
    return cmd

def nmap_scan(args, host):
    logging.info("Fingerprinting "+host)
    if os.geteuid() == 0:
        logging.info("Running nmap as root user")
    else:
        logging.info("Running nmap as non-root user")
    nmap_cmd = create_nmap_cmd(args)
    if args.verbosity >= 3:
        logging.debug('Enabled nmap debug logging...')
        nmap_cmd = nmap_cmd + ' -vvv -d --packet-trace --reason'
    elif args.verbosity == 2:
        nmap_cmd = nmap_cmd + ' -vv -d '
    cmdarr = [nmap_cmd + ' ' + host]
    try:
        logging.debug("NMAP command: " + cmdarr[0])
        out = subprocess.check_output(cmdarr, shell=True)
        out = out.decode(args.encoding)
    except subprocess.CalledProcessError:
        logging.error("Error running nmap command")
        return [] 

    logging.debug("NMAP output:\n" + out)
    dom = parseString(out)
    asset_data_list = []
    hosts = dom.getElementsByTagName("host")
    for h in hosts:
        addr = h.getElementsByTagName("address")[0]
        addr = addr.getAttribute('addr')
        hostname = addr
        harr = h.getElementsByTagName("hostname")
        if harr != None and len(harr) > 0:
            hostname = h.getElementsByTagName("hostname")[0]
            hostname = hostname.getAttribute('name')
            if hostname == 'linux':
                hostname = addr
        
        # Check for ports in use
        ports_in_use_dict = { }
        ports = h.getElementsByTagName("port")
        for port in ports:
            if not is_port_open(port):
                continue
            port_no = int(port.getAttribute('portid'))
            protocol = port.getAttribute('protocol')
            #service = port.getElementsByTagName('service')[0]
            service = port.getElementsByTagName('service')
            if service == None or len(service) == 0:
                continue
            service = service[0]
            service_product = None
            if service is not None:
                service_product = service.getAttribute('product')
                service = service.getAttribute('name')
            ports_in_use_dict[port_no] = { "port": str(port_no), "protocol": protocol, "service": service, "product": service_product}

        # SSH Port in use?
        ssh_port_is_open = True if ports_in_use_dict.get(22) is not None else False
        # HTTP(S) ports in use
        http_port_80_is_open = True if ports_in_use_dict.get(80) is not None else False
        http_port_8080_is_open = True if ports_in_use_dict.get(8080) is not None else False
        https_port_443_is_open = True if ports_in_use_dict.get(443) is not None else False
        https_port_8443_is_open = True if ports_in_use_dict.get(8443) is not None else False

        # check for cpes
        cpes = h.getElementsByTagName("cpe")
        products = []
        for c in cpes:
            # ignore cpes that are part of osclass section
            if c.parentNode.tagName == "osclass":
                continue
            cstr = c.firstChild.data
            carr = cstr.split(':')
            prodstr = carr[2] + ' ' + carr[3] + ' '
            if len(carr) >= 5:
                prodstr += carr[4]
            prodstr = prodstr.strip()
            prodstr = prodstr.replace('_',' ')
            if prodstr not in products:
                products.append(prodstr)

        ostype = get_os_type(h, products)

        # check for services
        services = h.getElementsByTagName("service")
        if services is not None:
            for s in services:
                if s.getAttribute('name') == 'snmp' and 'snmp' in args.services:
                    # use snmpwalk to find more
                    if not SNMPWALK:
                        logging.warn("snmpwalk command not found")
                        continue
                    # build the snmp walk command
                    cmd = build_snmp_walk_cmd(args, addr)
                    prod = get_snmp_oid_value(args, cmd, '1.3.6.1.2.1.1.1.0')
                    if prod:
                        logging.debug("SNMP sysDescr value:"+prod)
                        # handle Palo Alto Networks products
                        if 'Palo Alto Networks' in prod:
                            #products.append(prod)
                            # oid for PANOS version
                            panosver = get_snmp_oid_value(args, cmd, '1.3.6.1.4.1.25461.2.1.2.1.1')
                            prod = 'paloaltonetworks pan-os '+panosver
                            products.append(prod)
                            ostype = 'Palo Alto Networks'
                        elif 'Juniper Networks' in prod:
                            version_regex = re.compile(r'\s([0-9A-Za-z_\-]+(\.[0-9A-Za-z_\-]+)+)')
                            m = re.findall(version_regex, prod)
                            if m and len(m) > 0:
                                junosver = 'juniper junos ' + [x[0] for x in m][0]
                                products.append(junosver)
                            model = prod.split(',')[1].replace('Inc.','').strip()
                            if '[' in model and ']' in model:
                                model = model.split()[1].replace('[','').replace(']','').strip()
                            else:
                                model = model.split()[0].strip()
                            products.append('juniper '+model)
                            ostype = 'Juniper'
                        elif 'Canon' in prod:
                            model = prod.split()[1].strip()
                            products.append('Canon '+model)
                            os_type = 'Canon Printer'
                        elif 'Honeywell' in prod:
                            prod = prod.replace(';',' ')
                            products.append(prod)
                            os_type = 'Honeywell Printer'
                        elif 'Zebra' in prod:
                            prod = prod.split('/')[0].strip()
                            products.append(prod)
                            os_type = 'Zebra Printer'
                        elif 'Fortinet' in prod: 
                            fgosver = get_snmp_oid_value(args, cmd, '1.3.6.1.4.1.12356.101.4.1.1')
                            if fgosver: # fortios
                                fgosver = 'fortinet fortios '+fgosver.split(',')[0].replace('v','')
                                ostype = 'Fortinet'
                                products.append(fgosver)
                                fgmodel = get_snmp_oid_value(args, cmd, '1.3.6.1.4.1.12356.100.1.1.1')
                                if fgmodel:
                                    model_regex = re.compile(r'\s([A-Za-z]+([0-9]+[A-Za-z]))')
                                    m = re.findall(model_regex, fgmodel)
                                    if m and len(m) > 0:
                                        model = 'fortinet ' + [x[0] for x in m][0]
                                        products.append(model)
                        elif 'Aruba' in prod:
                            ostype = 'Aruba'
                            swmodel = prod.split()[2].split('-')[0]
                            version_regex = re.compile(r'([0-9]+\.[0-9]+\.[0-9]+)')
                            aos_version = None
                            m = re.findall(version_regex, prod)
                            if m and len(m) > 0:
                                aos_version = m[0]
                            modelnum = re.sub("[^\d]", "", swmodel)
                            if int(modelnum) < 6000:
                                if aos_version:
                                    products.append(swmodel+' arubaos-switch '+aos_version)
                                    products.append(swmodel+' firmware '+aos_version)
                            else:
                                if aos_version:
                                    products.append(swmodel+' aos-cx '+aos_version)
                        elif 'Cisco' in prod:
                            ostype = 'Cisco'
                            ver = ''
                            if 'Version' in prod:
                                ver = prod.split('Version')[1].split()[0].strip()
                                ver = ver.replace(',','')
                            os_prod = None 
                            if 'IOS' in prod:
                                if 'XE' in prod or 'IOS-XE' in prod:
                                    os_prod = 'cisco ios xe software '+ver
                                else:
                                    os_prod = 'cisco ios '+ver
                            elif 'NX-OS(tm)' in prod:
                                os_prod = 'cisco nx-os software '+ver
                                device = prod.split(',')[0]
                                if 'Nexus' in device:
                                    device = device.replace('Nexus','Nexus ')
                                    device = device.replace('NX-OS(tm)','')
                                    if len(device.split()) > 3:
                                        dlist = device.split()[:-1]
                                        device = " ".join(dlist)
                                    device = device + " series devices running nx-os software"
                                    products.append(device)
                            if os_prod:
                                os_prod = os_prod.strip()
                                products.append(os_prod)
                prod = s.getAttribute('product')
                if not prod:
                    continue
                ver = s.getAttribute('version')
                if not ver:
                    continue
                prod = prod + ' ' + ver
                if prod not in products:
                    products.append(prod)
        # check for script output
        scripts = h.getElementsByTagName("script")
        for s in scripts:
            if s.getAttribute('id') == 'http-wordpress-enum':
                wpout = s.getAttribute('output')
                if wpout != None:
                    wplist = wpout.splitlines()
                    for wp in wplist:
                        wp = wp.strip()
                        if wp == '':
                            continue
                        if wp.startswith('Search limited to'):
                            continue
                        if wp == 'plugins':
                            continue
                        if wp == 'themes':
                            continue
                        prodstr = 'wordpress plugin '+wp
                        if prodstr not in products:
                            products.append(prodstr)
            elif s.getAttribute('id') == 'tomcat-version':
                wpout = s.getAttribute('output')
                if wpout != None and wpout != '':
                    prodstr = 'apache tomcat ' + wpout.split(':')[1].strip()
                    if prodstr not in products:
                        products.append(prodstr)
            elif s.getAttribute('id') == 'http-server-header':
                wpout = s.getAttribute('output')
                if wpout != None and wpout != '':
                    prods = wpout.split(',')
                    for prod in prods:
                        prod = prod.replace('/',' ').strip()
                        prod = prod.replace('(',' ').strip()
                        prod = prod.replace(')',' ').strip()
                        if prod not in products:
                            products.append(prod)
            elif s.getAttribute('id') == 'http-php-version':
                wpout = s.getAttribute('output')
                if wpout != None and wpout != '':
                    if ':' not in wpout:
                        continue
                    prodstr = wpout.split(':')[1].strip()
                    prodstr = prodstr.split(',')[0]
                    prodstr = prodstr.split(' ')[0]
                    prodstr = prodstr.replace('PHP/','')
                    prodstr = prodstr.replace(',','')
                    prodstr = 'php '+prodstr
                    if prodstr not in products:
                        products.append(prodstr)
            elif s.getAttribute('id') == 'mirth-connect-version':
                wpout = s.getAttribute('output')
                if wpout != None and wpout != '':
                    prodstr = 'mirth connect ' + wpout.split(':')[1].strip()
                    if prodstr not in products:
                        products.append(prodstr)
            elif s.getAttribute('id') == 'connectwise-screenconnect-version':
                wpout = s.getAttribute('output')
                if wpout != None and wpout != '':
                    prodstr = 'connectwise screenconnect ' + wpout.split(':')[1].strip()
                    if prodstr not in products:
                        products.append(prodstr)
            elif s.getAttribute('id') == 'http-generator':
                wpout = s.getAttribute('output')
                if wpout != None and wpout != '':
                    prodstr = wpout.strip()
                    if prodstr not in products:
                        products.append(prodstr)
            elif s.getAttribute('id') == 'mysql-info':
                elems = s.getElementsByTagName('elem')
                for e in elems:
                    key = e.getAttribute('key')
                    if key and key == 'Version':
                        prodstr = 'mongodb '+e.firstChild.data
                        if prodstr not in products:
                            products.append(prodstr)
            elif s.getAttribute('id') == 'erldp-info':
                wpout = s.getAttribute('output')
                if wpout != None:
                    prodstr = 'erldp ' + wpout.split('version:')[1].split('node')[0].strip()
                    if prodstr not in products:
                        products.append(prodstr)
            elif s.getAttribute('id') == 'confluence-version':
                wpout = s.getAttribute('output')
                if wpout != None:
                    if 'atlassian.net' in hostname:
                        prodstr = 'atlassian confluence ' + wpout.split('version:')[1].strip()
                    else:
                        prodstr = 'atlassian confluence data center ' + wpout.split('version:')[1].strip()
                    if prodstr not in products:
                        products.append(prodstr)
            elif s.getAttribute('id') == 'jira-version':
                wpout = s.getAttribute('output')
                if wpout != None and 'version:' in wpout:
                    if 'atlassian.net' in hostname:
                        prodstr = 'atlassian jira ' + wpout.split('version:')[1].strip()
                    else:
                        prodstr = 'atlassian jira data center ' + wpout.split('version:')[1].strip()
                    if prodstr not in products:
                        products.append(prodstr)
            elif s.getAttribute('id') == 'ms-sql-info':
                tables = s.getElementsByTagName('table')
                for table in tables:
                    if table.getAttribute('key') =='Version':
                        elems = table.getElementsByTagName('elem')
                        prod = ''
                        version = ''
                        for elem in elems:
                            if elem.getAttribute('key') == 'name':
                                prod = elem.firstChild.nodeValue.strip()
                            if elem.getAttribute('key') == 'number':
                                version = elem.firstChild.nodeValue.strip()
                        prod = prod + ' ' + version
                        if prod not in products:
                            products.append(prod)
            elif s.getAttribute('id') == 'vmware-version':
                wpout = s.getAttribute('output')
                if wpout != None:
                    prod = wpout.split('Server version:')[1].split('Locale')[0].strip()
                    prod = prod.replace('Server ','')
                    prod = prod.replace('\n','')
                    prod = prod.replace('Build: ','build-')
                    if prod not in products:
                        products.append(prod)
            elif s.getAttribute('id') == 'zookeeper-info':
                wpout = s.getAttribute('output')
                if wpout != None and 'zookeeper.version=' in wpout:
                    prod = 'apache zookeeper ' + wpout.split('zookeeper.version=')[1].split('-')[0].strip()
                    if prod not in products:
                        products.append(prod)
            elif s.getAttribute('id') == 'hp-printers':
                wpout = s.getAttribute('output')
                if wpout != None and 'hp printer:' in wpout:
                    prod = wpout.split('hp printer:')[1].strip()
                    if prod not in products:
                        products.append(prod)
                    ostype = 'HP Printer'
            elif s.getAttribute('id') == 'arecont-cctv':
                wpout = s.getAttribute('output')
                if wpout != None:
                    if wpout not in products:
                        products.append(wpout)
                    ostype = 'Arecont Vision'
            elif s.getAttribute('id') == 'axis-cctv':
                wpout = s.getAttribute('output')
                if wpout != None:
                    if wpout not in products:
                        products.append(wpout)
                    ostype = 'Axis Communications'
            elif s.getAttribute('id') == 'hanwhavision-cctv':
                wpout = s.getAttribute('output')
                if wpout != None:
                    wpout = wpout.replace('Hanwha Vision','hanwhavision').strip()
                    if wpout not in products:
                        products.append(wpout)
                    ostype = 'Hanwha Vision'
            elif s.getAttribute('id') == 'goanywhere_version':
                wpout = s.getAttribute('output')
                if wpout != None and wpout not in products:
                    products.append(wpout)

        os_name_tag = None
        smb_os_name = get_smb_os(h)
        # get the product name associated with the best os class
        best_os_class = get_best_os_class(h) 
        if best_os_class:
            if ostype == "Linux":
                cpe = best_os_class.getElementsByTagName("cpe")[0]
                cstr = cpe.firstChild.data
                carr = cstr.split(':')
                prodstr = carr[2] + ' ' + carr[3] + ' '
                if len(carr) >= 5:
                    prodstr += carr[4]
                prodstr = prodstr.strip()
                prodstr = prodstr.replace('_',' ')
                if "." not in prodstr:
                    # At times version is incomplete i.e. like a single digit like "linux linux kernel 4" and this does not lend itself nicely for matching. Hence add the ".0" to complete it.
                    prodstr = prodstr + ".0"
                if prodstr not in products:
                    products.append(prodstr)
            elif ostype == "Windows":
                if smb_os_name:
                    products.append(smb_os_name)
                    os_name_tag = smb_os_name
                else:
                    os_name = best_os_class.parentNode.getAttribute("name")
                    # At times there can be multiple os_names like "Microsoft Windows 7, Windows Server 2012, or Windows 8.1 Update 1", then take the first one
                    if "," in os_name and " or " in os_name:
                        comma_position = os_name.find(",")
                        or_position = os_name.find(" or ")
                        if comma_position < or_position:
                            os_name = os_name.split(',')[0]
                        else:
                            os_name = os_name.split(" or ")[0]
                    elif "," in os_name:
                        os_name = os_name.split(",")[0]
                    elif " or " in os_name:
                        os_name = os_name.split(' or ')[0]
                    products.append(os_name)
                    os_name_tag = os_name

        # clean up the list of products
        clean_products = []
        for p in products:
            # remove bare references to windows and linux kernels without versions
            if p.lower() == 'microsoft windows' or p.lower() == 'linux linux kernel':
                continue
            if ostype == "Windows":
                # if smb os name is present then remove all references to other microsoft windows
                if smb_os_name and p.lower().startswith('microsoft windows'):
                    continue
                # remove references to linux kernel in windows asset 
                if p.lower().startswith('linux linux kernel'):
                    continue
            if ostype == "Linux":
                # remove references to linux kernel in windows asset 
                if p.lower().startswith('microsoft windows'):
                    continue
            clean_products.append(p)
        products = clean_products

        asset_data = {}
        asset_data['id'] = addr 
        asset_data['name'] = hostname 
        asset_data['type'] = ostype
        asset_data['owner'] = args.handle
        asset_data['products'] = products
        asset_tags = ["DISCOVERY_TYPE:Unauthenticated"]
        if os_name_tag:
            asset_tags.append("OS_RELEASE:" + os_name_tag)
            asset_tags.append(os_name_tag)
        asset_data['tags'] = asset_tags

        if 'printers' not in args.services and 'cctv' not in args.services and 'snmp' not in args.services:
            if len(ports_in_use_dict) > 0:
                asset_data['config_issues'] = create_open_ports_issues(ports_in_use_dict, addr)
            if args.no_ssh_audit == False and ssh_port_is_open:
                ssh_issues = linux.run_ssh_audit(args, addr, addr)
                if len(ssh_issues) != 0:
                    asset_data['tags'].append('SSH Audit')
                asset_data['config_issues'] = asset_data['config_issues'] + ssh_issues if 'config_issues' in asset_data else ssh_issues

        if asset_data['type'] == "Other" and len(asset_data['products']) == 0 and ('config_issues' not in asset_data or len(asset_data['config_issues'])==0):
            # skip any discovered assets which have asset type as "Other" and no products and no config_issues
            logging.info("Fingerprinting did not yield any results")
            continue
        asset_data_list.append(asset_data)

    return asset_data_list

def get_inventory(args): 
    if not nmap_exists():
        logging.error('nmap CLI not found')
        return None

    if args.hosts == None:
        args.hosts = get_private_ip_cidrs()
    else:
        args.hosts = args.hosts.split(',')
    assets = []
    for host in args.hosts:
        assets.extend(nmap_scan(args, host))
    return assets 
